import argparse

_field_name_map_ = {
    'del': 'Q(del)',
    'err_ont': 'Q(acc)',
    'ins': 'Q(ins)',
    'iden': 'Q(iden)',
    'sub': 'Q(sub)',
}

_cols_ = ['name', 'ref', 'ref_cover', 'Q(acc)', 'Q(iden)', 'Q(del)', 'Q(ins)', 'Q(sub)']

_format_str_ = '\t'.join(['{' + c + '}' for c in _cols_])

_median_col_ind_ = {True: 3, False: 1}


def get_ref_cover(lines, ref):
    for l in lines:
        s = l.split(' ')
        if s[0] == ref:
            return s[1]


def extract_vals(vals, name_map, col_ind):
    results = {}
    for line in vals:
        s = line.split()
        name = s[0]
        val = s[col_ind]
        if name in name_map:
            results[name_map[name]] = val
    return results


def main():
    parser = argparse.ArgumentParser(
        description='Extract Q scores from summary_from_stats output',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('summaries', nargs='+', help='*summ.txt created by summary_from_stats')
    parser.add_argument('--median', help='Use median. If false, use mean.', action='store_true', default=False)
    parser.add_argument('--ref', default=None, help='process single ref, rather than overall result')
    args = parser.parse_args()
    qscore_from_summary(args.summaries, args.median, args.ref)


def qscore_from_summary(summaries, median, ref):
    if ref is None:
        match = '#  Q Scores'
    else:
        match = '# {} Q Scores'.format(ref)

    col_ind = _median_col_ind_[median]

    print('\t'.join(_cols_))

    row_period = len(_field_name_map_) + 3

    results = []
    for f in sorted(summaries):
        with open(f) as fh:
            lines = [l.strip() for l in fh.readlines()]
        if match not in lines:
            continue
        ind = lines.index(match)
        vals = lines[ind + 2: ind + row_period]
        data = extract_vals(vals, _field_name_map_, col_ind)
        data['name'] = f
        if ref is None:
            data['ref_cover'] = 'NA'
            data['ref'] = 'all'
        else:
            data['ref_cover'] = get_ref_cover(lines, ref)
            data['ref'] = ref
        print(_format_str_.format(**data))
        results.append(data)

    return results


if __name__ == "__main__":
    main()



import argparse
from concurrent.futures import ProcessPoolExecutor
import functools
import logging
import multiprocessing
import os

from intervaltree import IntervalTree, Interval
import numpy as np
import pysam


from pomoxis.util import parse_regions, Region, filter_args, filter_read, write_bam
from pomoxis.coverage_from_bam import coverage_summary_of_region


def main():
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    parser = argparse.ArgumentParser(
        prog='subsample_bam',
        description='Subsample a bam to uniform or proportional depth',
        parents=[filter_args()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bam',
        help='input bam file.')
    parser.add_argument('depth', nargs='+', type=int,
        help='Target depth.')
    parser.add_argument('-o', '--output_prefix', default='sub_sampled',
        help='Output prefix')
    parser.add_argument('-r', '--regions', nargs='+',
        help='Only process given regions.')
    parser.add_argument('-p', '--profile', type=int, default=1000,
        help='Stride in genomic coordinates for depth profile.')
    parser.add_argument('-t', '--threads', type=int, default=-1,
        help='Number of threads to use.')
    parser.add_argument('--force_low_depth', action='store_true',
        help='Force saving reads mapped to a sequence with coverage below the expected value.')

    eparser = parser.add_mutually_exclusive_group()
    eparser.add_argument('--any_fail', action='store_true',
        help='Exit with an error if any region has insufficient coverage.')
    eparser.add_argument('--all_fail', action='store_true',
        help='Exit with an error if all regions have insufficient coverage.')

    uparser = parser.add_argument_group('Uniform sampling options')
    uparser.add_argument('-x', '--patience', default=5, type=int,
        help='Maximum iterations with no change in median coverage before aborting.')
    uparser.add_argument('-s', '--stride', type=int, default=1000,
        help='Stride in genomic coordinates when searching for new reads. Smaller can lead to more compact pileup.')

    pparser = parser.add_argument_group('Proportional sampling options')
    pparser.add_argument('-P', '--proportional', default=False, action='store_true',
        help='Activate proportional sampling, thus keeping depth variations of the pileup.')
    pparser.add_argument('-S', '--seed', default=None, type=int,
        help='Random seed for proportional downsampling of reads.')

    args = parser.parse_args()
    if args.threads == -1:
        args.threads = multiprocessing.cpu_count()

    with pysam.AlignmentFile(args.bam) as bam:
        ref_lengths = dict(zip(bam.references, bam.lengths))

        if args.regions is not None:
            regions = parse_regions(args.regions, ref_lengths=ref_lengths)
        else:
            regions = [Region(ref_name=r, start=0, end=ref_lengths[r]) for r in bam.references]

    if args.proportional:
        worker = functools.partial(subsample_region_proportionally, args=args)
    else:
        worker = functools.partial(subsample_region_uniformly, args=args)

    enough_depth = []
    with ProcessPoolExecutor(max_workers=args.threads) as executor:
        for res in executor.map(worker, regions):
            enough_depth.append(res)

    if args.any_fail and not all(enough_depth):
        raise RuntimeError('Insufficient read coverage for one or more requested regions.')
    if args.all_fail and all([not x for x in enough_depth]):
        raise RuntimeError('Insufficient read coverage for all requested regions.')


def subsample_region_proportionally(region, args):
    logger = logging.getLogger(region.ref_name)
    coverage_summary = coverage_summary_of_region(
        region, args.bam, args.stride, functools.partial(filter_read, args=args, logger=logger))
    col = 'depth_{}'.format(args.orientation) if args.orientation is not None else 'depth'
    median_coverage = coverage_summary[col].T['50%']
    logger.info(f'Median coverage {median_coverage}')
    with pysam.AlignmentFile(args.bam) as bam:
        def _read_iter():
            for r in bam.fetch(region.ref_name, region.start, region.end):
                if not filter_read(r, args, logger):
                    yield r
        # Prepare iterator
        reads = _read_iter()
        # query names cannot be longer than 251
        dtype = [('name', 'U251'), ('start', int), ('end',  int)]
        read_data = np.fromiter(
            ((r.query_name, r.reference_start, r.reference_end) for r in reads),
            dtype=dtype
        )
        if len(read_data) == 0:
            logger.warn(f'No primary reads found in {region.ref_name}.')
            found_enough_depth = False
            return found_enough_depth

    targets = sorted(args.depth)
    found_enough_depth = True

    coverage = np.zeros(region.end - region.start, dtype=np.uint16)
    if args.seed is not None:
        np.random.seed(args.seed)

    for target in targets:
        if target > median_coverage:
            found_enough_depth = False
            if not args.force_low_depth:
                msg = 'Target depth {} exceeds median coverage, limiting to {} and skipping higher depths.'
                logger.warn(msg.format(target, int(median_coverage)))
                break
        fraction = target / median_coverage
        if fraction <= 1:
            n_reads = int(round(fraction * len(read_data), 0))
            target_reads = np.random.choice(read_data, n_reads, replace=False)
            logger.info('Processing {}X: {} reads ({:.2f} %).'.format(target, n_reads, 100 * fraction))
        else:
            n_reads = len(read_data)
            target_reads = read_data
            found_enough_depth = False
            logger.warn('Insufficient coverage: {} reads ({}X).'.format(n_reads, median_coverage))
        if found_enough_depth:
            prefix = '{}_{}X'.format(args.output_prefix, target)
        else:
            prefix = '{}_{}X'.format(args.output_prefix, int(median_coverage))
        if n_reads > 0:
            write_bam(args.bam, prefix, region, target_reads['name'], keep_supplementary=args.keep_supplementary)
        coverage.fill(0.0)  # reset coverage for each target depth
        for read in target_reads:
            coverage[read['start'] - region.start:read['end'] - region.start] += 1
        _write_coverage(prefix, region, coverage, args.profile)

    return found_enough_depth


def subsample_region_uniformly(region, args):
    logger = logging.getLogger(region.ref_name)
    logger.info("Building interval tree.")
    tree = IntervalTree()
    with pysam.AlignmentFile(args.bam) as bam:
        ref_lengths = dict(zip(bam.references, bam.lengths))
        for r in bam.fetch(region.ref_name, region.start, region.end):
            if filter_read(r, args, logger):
                continue
            # trim reads to region
            tree.add(Interval(
                max(r.reference_start, region.start), min(r.reference_end, region.end),
                r.query_name))

    logger.info('Starting pileup.')
    coverage = np.zeros(region.end - region.start, dtype=np.uint16)
    reads = set()
    n_reads = 0
    iteration = 0
    it_no_change = 0
    last_depth = 0
    targets = iter(sorted(args.depth))
    target = next(targets)
    found_enough_depth = True
    while True:
        cursor = 0
        while cursor < ref_lengths[region.ref_name]:
            read = _nearest_overlapping_point(tree, cursor)
            if read is None:
                cursor += args.stride
                continue
            else:
                reads.add(read.data)
                cursor = read.end
                coverage[read.begin - region.start:read.end - region.start] += 1
                tree.remove(read)

        iteration += 1
        median_depth = np.median(coverage)
        stdv_depth = np.std(coverage)
        logger.debug(u'Iteration {}. reads: {}, depth: {:.0f}X (\u00B1{:.1f}).'.format(
            iteration, len(reads), median_depth, stdv_depth))
        # output when we hit a target
        if median_depth >= target or not found_enough_depth:
            if found_enough_depth:
                msg = "Hit target depth {}."
                logger.info(msg.format(target))
                prefix = '{}_{}X'.format(args.output_prefix, target)
                write_bam(args.bam, prefix, region, reads, keep_supplementary=args.keep_supplementary)
                _write_coverage(prefix, region, coverage, args.profile)
                try:
                    target = next(targets)
                except StopIteration:
                    break
            else:
                msg = "Target depth {} exceeds median coverage, limiting to {} and skipping higher depths."
                logger.warn(msg.format(target, int(median_depth)))
                prefix = '{}_{}X'.format(args.output_prefix, int(median_depth))
                if median_depth > 0:
                    write_bam(args.bam, prefix, region, reads, keep_supplementary=args.keep_supplementary)
                    _write_coverage(prefix, region, coverage, args.profile)
                break
        # exit if nothing happened this iteration
        if n_reads == len(reads):
            found_enough_depth = False
            if not args.force_low_depth:
                logger.warn("No reads added, finishing pileup.")
                break
        n_reads = len(reads)
        # or if no change in depth
        if median_depth == last_depth:
            it_no_change += 1
            if it_no_change == args.patience:
                found_enough_depth = False
                if not args.force_low_depth:
                    logging.warn("Coverage not increased for {} iterations, finishing pileup.".format(
                        args.patience
                    ))
                    break
        else:
            it_no_change == 0
        last_depth = median_depth
    return found_enough_depth


def _nearest_overlapping_point(src, point):
    """Find the interval with the closest start point to a given point.

    :param src: IntervalTree instance.
    :param point: query point.

    :returns: Interval instance of interval with closest start.

    """
    items = src.at(point)
    if len(items) == 0:
        return None
    items = sorted(items, key=lambda x: x.end - x.begin, reverse=True)
    items.sort(key=lambda x: abs(x.begin - point))
    return items[0]


def _write_coverage(prefix, region, coverage, profile):
    # depth profile
    output = '{}_{}.depth'.format(prefix, region.ref_name)
    end = profile * (len(coverage) // profile)
    cov_blocks = coverage[0:end].reshape(-1, profile)
    depth_profile = np.mean(cov_blocks, axis=1, dtype=np.uint32)
    start = region.start + profile // 2
    positions = (start + profile * x for x in range(len(depth_profile)))
    with open(output, 'w') as fh:
        fh.write("position\tdepth\n")
        for pos, depth in zip(positions, depth_profile):
            fh.write("{}\t{}\n".format(pos, depth))


if __name__ == '__main__':
    main()



"""
Tabulate some simple alignment stats from sam.
"""

import argparse
import sys
import numpy as np
import pandas as pd
import warnings
from collections import OrderedDict
import re

parser = argparse.ArgumentParser(
        description="""Summarise output of `stats_from_bam`.""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('-i', '--input', type=argparse.FileType('r'), default=sys.stdin)
parser.add_argument('-o', '--output', type=argparse.FileType('w'), default=sys.stderr,
                    help='Output summary to file instead of stderr.')
parser.add_argument('-p', '--percentiles', type=int, nargs='+', default=(10,50,90),
                    help='Percentiles for summary.')
parser.add_argument('-pr', '--per_reference', action='store_true',
                    help='Also output a summary for each reference.')
parser.add_argument('-a', '--accumulate', type=int, default=1,
                    help='Accumulate chunks in the summary.')

def qscore(d):
    with warnings.catch_warnings():
        # some values might be zero
        warnings.simplefilter("ignore")
        q = -10 * np.log10(d)
    return q

def summarise_stats(d, percentiles=(10, 50, 90), accumulate=1):

    def sum_every_n(x):
        return np.add.reduceat(np.array(x), np.arange(0, len(x), accumulate))

    s = []
    def summarise_stat(name, f):
        # per alignment errors from which to calculate metrics, bounds etc.
        r = OrderedDict()
        r['name'] = name
        r['mean'] = f[0](d)
        vals = f[1](d)
        for p, pval in zip(percentiles, np.percentile(vals, percentiles)):
            r['q{}'.format(p)] = pval
        return r

    # create functions to get overall length-weighted averages
    # and per-alignment errors from which to calculate percentiles
    f = OrderedDict()

    if accumulate > 1:
        per_align_func = sum_every_n
    else:
        per_align_func = lambda x: x

    f['err_ont'] = (lambda d: np.sum(d['sub'] + d['ins'] + d['del']) / np.sum(d['length']), # weighted
                    lambda d: per_align_func(d['sub'] + d['ins'] + d['del']) / per_align_func(d['length'])) # per n alignments
    f['err_bal'] = (lambda d: np.sum(d['sub'] + d['ins'] + d['del']) / np.sum(d['aligned_ref_len']),
                    lambda d: per_align_func(d['sub'] + d['ins'] + d['del']) / per_align_func(d['aligned_ref_len']))
    f['iden'] = (lambda d: np.sum(d['sub']) / (np.sum(d['aligned_ref_len'] - d['del'])),
                 lambda d: per_align_func(d['sub']) / per_align_func(d['aligned_ref_len'] - d['del']))
    f['del'] = (lambda d: np.sum(d['del']) / (np.sum(d['aligned_ref_len'])),
               lambda d: per_align_func(d['del']) / per_align_func(d['aligned_ref_len']))
    f['ins'] = (lambda d: np.sum(d['ins']) / (np.sum(d['aligned_ref_len'])),
               lambda d: per_align_func(d['ins']) / per_align_func(d['aligned_ref_len']))
    f['sub'] = (lambda d: np.sum(d['sub']) / (np.sum(d['aligned_ref_len'])),
               lambda d: per_align_func(d['sub']) / per_align_func(d['aligned_ref_len']))

    for name, f in f.items():
        s.append(summarise_stat(name, f))

    return s

def main(arguments=None):
    args = parser.parse_args(arguments)
    n=args.accumulate

    if n<1:
        raise ValueError('accumulate should be at least 1.')

    stats = pd.read_csv(args.input, sep='\t')
    l=len(stats)

    if l == 0:
        raise ValueError('No alignments stats found.')

    # Warn if there are not many chunks to produce meaningful accumulated summary
    if n>1 and l<10*n:
        sys.stdout.write("WARNING! Accumulating "+str(l)+" chunks in groups of "+str(n)+" might not produce meaningful summary stats.\n")

    # Sort by query chromosome and chunk
    if n>1:
        qchr=[re.sub("_chunk.+","",s) for s in stats['name']]
        chunk=[re.sub(".+_chunk","",s) for s in stats['name']]
        # If chunk numbers are present convert to integers for sorting
        if all([r.isdigit() for r in chunk]):
            chunk=[int(r) for r in chunk]
        stats=stats.assign(qchr=qchr,chunk=chunk)
        stats.sort_values(by=["qchr","chunk"],inplace=True)

    def output_summary(stats, prefix=''):
        to_str_opts = {'buf': args.output, 'col_space': 8, 'index':False, 'justify': 'center'}
        errors = pd.DataFrame(summarise_stats(stats, percentiles=args.percentiles, accumulate=n))

        pcnt_formatter = {}
        for col in errors.columns:
            if np.issubdtype(errors[col].dtype, np.number):
                pcnt_formatter[col] = '{:5.3%}'.format
            else:
                pcnt_formatter[col] = '{}'.format
        args.output.write('\n\n# {} Percentage Errors\n'.format(prefix))
        errors.to_string(formatters=pcnt_formatter, **to_str_opts)

        args.output.write('\n\n# {} Q Scores\n'.format(prefix))
        for col in errors.columns:
            if np.issubdtype(errors[col].dtype, np.number):
                errors[col] = qscore(errors[col])
        errors.to_string(float_format='%5.2f', **to_str_opts)

    output_summary(stats)

    args.output.write('\n\n# Ref Coverage\n')
    for ref, grp in stats.groupby('ref'):
        args.output.write('{} {:5.3f}\n'.format(ref, grp['ref_coverage'].sum()))

    if args.per_reference:
        for ref, grp in stats.groupby('ref'):
            output_summary(grp, prefix=ref)

    args.output.write('\n')


if __name__ == '__main__':
    main()



import argparse
from collections import defaultdict, Counter, namedtuple
import concurrent.futures
from functools import partial
import itertools
import logging
import shutil
from math import ceil
from operator import attrgetter
import os
import pickle
import re
import unittest
import warnings

import matplotlib; matplotlib.use('Agg', force=True)  # enforce non-interactive backend
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import pysam

from pomoxis.util import get_trimmed_pairs, intervaltrees_from_bed

AlignSeg = namedtuple('AlignSeg', ('rname', 'qname', 'pairs', 'rlen'))
Error = namedtuple('Error', ('rp', 'rname', 'qp', 'qname', 'ref', 'match', 'read', 'counts', 'klass', 'aggr_klass'))
Context = namedtuple('Context', ('p_i', 'qb', 'rb'))

_match_ = ' '
_indel_ = ':'
_sub_ = '|'

_sep_ = '\t'

_indel_sizes_=(5,10,100)
_error_groups_ = [
                  ('HP sub swp', lambda x: 'HP sub swp' in x),
                  ('HP sub ext', lambda x: 'HP sub ext' in x),
                  ('HP sub trc', lambda x: 'HP sub trc' in x),
                  ('sub HP split', lambda x: 'sub HP split' in x),
                  ('ins HP split', lambda x: 'ins HP split' in x),
                  ('sub HP join', lambda x: 'sub HP join' in x),
                  ('del HP join', lambda x: 'del HP join' in x),
                  ('multi ins >', lambda x: bool(re.search('multi .*ins >', x))),
                  ('multi ins <', lambda x: bool(re.search('multi .*ins <', x))),
                  ('multi del >', lambda x: bool(re.search('multi .*del >', x))),
                  ('multi del <', lambda x: bool(re.search('multi .*del <', x))),
                  ('HP del', lambda x: 'HP del' in x),
                  ('HP ins', lambda x: 'HP ins' in x),
                  ('HP sub', lambda x: 'HP sub' in x),
                  ('fwd repeat ins', lambda x: 'fwd repeat ins' in x),
                  ('rev repeat ins', lambda x: 'rev repeat ins' in x),
                  ('fwd repeat del', lambda x: 'fwd repeat del' in x),
                  ('rev repeat del', lambda x: 'rev repeat del' in x),
                  ('ins', lambda x: 'ins' in x),
                  ('del', lambda x: 'del' in x),
                  ('sub', lambda x: 'sub' in x),
]

def get_errors(aln, tree=None):
    """Find positions of errors in an aligment.

    :param aln: iterable of `AlignPos` objects.
    :param bed_file: path to .bed file of regions to include in analysis.
    :param tree: `ncls.NCLS` object of regions to analyse.
    :returns: ( [(ri, qi, 'error_type', last_ri, last_qi)], aligned_ref_len)
        ri, qi: ref and query positions
        error_type: 'D', 'I' or 'S'
        last_ri, last_qi: ref and query positions of the last match
        aligned_ref_len: total aligned reference length (taking account of masking tree)
    """

    err = []
    last_qi = None
    last_ri = None
    pos = None
    n_masked = 0
    aligned_ref_len = 0
    match = 0
    for (qi, qb, ri, rb) in aln:
        if tree is not None:
            pos = ri if ri is not None else pos
            if not tree.has_overlap(pos, pos + 1) or (ri is None and not tree.has_overlap(pos + 1, pos + 2)):
                # if ri is None, we are in an insertion, check if pos + 1 overlaps
                # (ref position of ins is arbitrary)
                # print('Skipping ref {}:{}'.format(read.reference_name, pos))
                n_masked += 1
                continue
        if qi is None:  # deletion
            last_ri = ri  # ri will not be None
            err.append((ri, qi, 'D', (last_ri, last_qi)))
            aligned_ref_len += 1
        elif ri is None:
            last_qi = qi  # qi will not be None
            err.append((ri, qi, 'I', (last_ri, last_qi)))
        else:
            last_qi = qi
            last_ri = ri
            if qb != rb:
                err.append((ri, qi, 'S', (last_ri, last_qi)))
            aligned_ref_len += 1
            match += 1

    if match == 0:
        # no matches within bed regions - all bed ref positions were deleted.
        # skip this alignment.
        return None
    
    return err, aligned_ref_len, n_masked


def rle(it):
    """Calculate a run length encoding (rle), of an input vector.

    :param it: iterable.
    :returns: structured array with fields `start`, `length`, and `value`.
    """
    val_dtype = np.array(it[0]).dtype
    dtype = [('length', int), ('start', int), ('value', val_dtype)]

    def _gen():
        start = 0
        for key, group in itertools.groupby(it):
            length = sum(1 for x in group)
            yield length, start, key
            start += length
    return np.fromiter(_gen(), dtype=dtype)


def get_run(i, runs):
    """Find run to which the i'th element belongs.

    :param i: int, element index wihin input to `rle`.
    :returns: int, element index within runs to which i belongs.
    """
    ends = runs['start'] + runs['length']
    start_i = min(np.searchsorted(runs['start'], i), len(runs) - 1)
    end_i = min(np.searchsorted(ends, i), len(runs) - 1)
    start = runs['start'][start_i]
    end = ends[end_i] - 1
    run_i = start_i if np.argmin(np.abs([start - i, end - i])) == 0 else end_i

    return run_i


def _get_context_bounds(p, aln, search_by_q, offset):
    """Find start and en d of context.

    In the simplest case this will be p-offset:p+offset, but we adjust
    for boundaries and to ensure we don't start/end on an error or within a HP.

    :param p: int, position (ref position, or query position)
    :param aln: iterable of `AlignPos` objects.
    :param search_by_q: bool, whether to search by query position (typically done if qi is None).
    :returns: (int start index, int end index, int index of p within aln[start:end])
    """
    if search_by_q:
        pos = [x.qpos for x in aln]
    else:
        pos = [x.rpos for x in aln]


    offset_tmp = offset
    start_p = max(p - offset, pos[0])
    end_p = min(p + offset, pos[-1])
    s = pos.index(start_p)
    e = pos.index(end_p)
    # extend context until it does not start/end with an error or in a HP.
    while True:
        s_is_match = aln[s].qbase == aln[s].rbase
        sq_not_hp = aln[s+1].qbase != aln[s].qbase
        sr_not_hp = aln[s+1].rbase != aln[s].rbase
        if (s_is_match and sq_not_hp and sr_not_hp) or s == 0:
            break
        else:
            s -= 1
    while True:
        e_is_match = aln[e-1].qbase == aln[e-1].rbase
        eq_not_hp = aln[e-1].qbase != aln[e-2].qbase
        er_not_hp = aln[e-1].rbase != aln[e-2].rbase
        if (e_is_match and eq_not_hp and er_not_hp) or e == len(aln):
            break
        else:
            e += 1

    p_i = pos.index(p)

    return s, e, p_i - s


def are_adjacent(inds):
    """"Check if all int indices in the interable are consecutive.

    :param inds: iterable of ints
    :returns: bool
    """
    if len(inds) == 1:
        adjacent = True
    else:
        adjacent = False
        i = inds[0]
        for n in inds[1:]:
            if n - i != 1:
                break
            i = n
        else:
            adjacent = True
    return adjacent


def is_in_hp(seq, p_i):
    n_indels_fwd = len([x for x in itertools.takewhile(lambda x: x=='-', seq[p_i+1:])])
    n_indels_rev = len([x for x in itertools.takewhile(lambda x: x=='-', seq[:p_i][::-1])])
    if seq[p_i] == '-':
        p_is_hp = seq[p_i - n_indels_rev - 1].upper() == seq[p_i + n_indels_fwd + 1].upper()
    else:
        p_is_hp = (seq[p_i].upper() == seq[p_i - 1 - n_indels_rev].upper() or
                    seq[p_i].upper() == seq[p_i + 1 + n_indels_fwd].upper())
    return p_is_hp


def classify_hp_sub(p_i, adjacent, errors, match_line, rb_runs,
                    qb_runs, qp_is_hp, rp_is_hp):
    hp_kls = None
    # sub errors
    # 'swap', e.g. r TTCCC -> q TTTCC or  r CCTTCCC -> q CTTTTCC
    # 'split', e.g. r TTTTTT -> q TTCCTT or r TTTTTT -> q TCTTCT
    # 'join', e.g. r TTCCTT -> TTTTTT
    # 'trunc', e.g. r TTTTTT -> q TTTTTC, r TTTTTT -> q CTTTTC,
    #               r TT -> q TC, r TT -> q GC
    # 'ext', e.g. r TTTTTC -> q TTTTTT
    rp_run_ind = get_run(p_i, rb_runs)
    qp_run_ind = get_run(p_i, qb_runs)
    rp_run = rb_runs[rp_run_ind]
    qp_run = qb_runs[qp_run_ind]
    sub_rb_inds = [get_run(i, rb_runs) for i in errors['sub']]
    sub_qb_inds = [get_run(i, qb_runs) for i in errors['sub']]
    # find query runs ref position run (if if ref is HP, get q over extent)
    q_run_inds_in_r_run = set([get_run(i, qb_runs) for i in range(rp_run['start'], rp_run['start'] + rp_run['length'])])
    q_runs_in_r_run = qb_runs[list(q_run_inds_in_r_run)]
    r_run_inds_in_q_run = set([get_run(i, rb_runs) for i in range(qp_run['start'], qp_run['start'] + qp_run['length'])])
    r_runs_in_q_run = rb_runs[list(r_run_inds_in_q_run)]
    cols = ['value', 'length']
    if qp_run[cols] == rp_run[cols]:
        # this is a false alarm, though some HP in the context has changed,
        # it was not as this position
        return hp_kls
    # if more than 1 query run is in ref run, we have a split / trunc
    # swap is a special case of truncation in which another HP has gained a base
    if len(q_run_inds_in_r_run) > 1:

        # all subs should belong to same RHP, otherwise we have a mess.
        if all([i == rp_run_ind for i in sub_rb_inds]):
            hp_start_ind = rp_run['start']
            hp_end_ind = hp_start_ind + rp_run['length'] - 1
            match_runs = rle(match_line)
            match_sub_inds = [get_run(i, match_runs) for i in errors['sub']]
            # if subs are not flanking we have a split
            if min(errors['sub']) > hp_start_ind and max(errors['sub']) < hp_end_ind:
                hp_kls = 'sub HP split ({}{})'.format(rp_run['value'], rp_run['length'])

            # if (possibly multiple adjacent) subs are all flanking the HP
            elif set(match_sub_inds).issubset([get_run(hp_start_ind, match_runs),
                                                get_run(hp_end_ind, match_runs)]):
                # if query is a HP, this is a swap
                if qp_run['length'] > 1:
                    hp_kls = 'HP sub swp ({}{}->{}{},{})'.format(rp_run['value'], rp_run['length'],
                                                                qp_run['value'], qp_run['length'] - len(errors['sub']), len(errors['sub']))
                else:
                    hp_kls = 'flk HP sub trc ({}{}->{})'.format(rp_run['value'], rp_run['length'],
                                                        rp_run['length'] - len(errors['sub']))
            else:
                hp_kls = 'complex HP sub trc ({}{})'.format(rp_run['value'], rp_run['length'])
        else:
            hp_kls = 'messy HP sub trc'

    # if more than 1 ref run is in query run, we have a join or extension
    # swap is a special case of extension in which another HP has lost a base
    elif len(r_run_inds_in_q_run) > 1:
        # all subs should belong to same QHP, else we have a mess
        if all([i == qp_run_ind for i in sub_qb_inds]):
            hp_start_ind = qp_run['start']
            hp_end_ind = hp_start_ind + qp_run['length'] - 1
            match_runs = rle(match_line)
            match_sub_inds = [get_run(i, match_runs) for i in errors['sub']]
            # if subs are not flanking we have a join
            if min(errors['sub']) > hp_start_ind and max(errors['sub']) < hp_end_ind:
                hp_kls = 'sub HP join ({}{})'.format(qp_run['value'], qp_run['length'])

            # if (possibly multiple adjacent) subs are all flanking the HP
            elif set(match_sub_inds).issubset([get_run(hp_start_ind, match_runs),
                                                get_run(hp_end_ind, match_runs)]):
                # if ref is a HP, this is a swap
                if rp_run['length'] > 1:
                    hp_kls = 'HP sub swp ({}{}->{}{},{})'.format(rp_run['value'], rp_run['length'],
                                                                qp_run['value'], qp_run['length'] - len(errors['sub']), len(errors['sub']))
                else:
                    hp_kls = 'flk HP sub ext ({}{}->{})'.format(qp_run['value'], qp_run['length']- len(errors['sub']), qp_run['length'])
            else:
                hp_kls = 'complex HP sub ext ({}{})'.format(qp_run['value'], qp_run['length'])
        else:
            hp_kls = 'messy HP sub ext'

    # if we just have 1 run of each, we might have a complete sub of the HP
    # e.g. TT->CC
    elif len(r_run_inds_in_q_run) == 1 and len(q_run_inds_in_r_run) == 1:
        if rp_run['value'] != qp_run['value']:
            hp_kls = 'HP sub swp ({}{}->{}{},{})'.format(rp_run['value'], rp_run['length'],
                                                          qp_run['value'], 0, qp_run['length'])

    return hp_kls


def classify_hp_indel(p_i, key, errors, runs1, seq2):
    """Look for a specific kind of HP indel that splits or joins two HPs

    :param p_i: int, index of error
    :param key: key of error type within errors (should be 'ins' or 'del')
    :param errors: dict of error positions
    :runs1: np.ndarray, rle encoding of sequence1 (query if deletions
           join two HPs, or ref if insertions split a HP
    :seq2: iterable of str of sequence2 (ref if deletions join two HPs,
           or query if insertions split a HP.
    :returns: str classification or None
    """
    # look for deletions in query HPs which indicate two ref HPs are joined.
    # or look for insertions in ref HPs which split them
    assert key in ['ins', 'del']
    _type_ = {'ins': 'split', 'del': 'join'}

    hp_kls = None
    p_run_ind = get_run(p_i, runs1)

    # if ref/query is not a HP, return
    if runs1[p_run_ind - 1]['value'] != runs1[p_run_ind + 1]['value']:
        logging.debug('Not a HP {} {}'.format(runs1[p_run_ind - 1]['value'],
                                              runs1[p_run_ind + 1]['value']))
        return hp_kls

    # if we have multiple non-adjacent indels, we could have
    # HP which should be split into >2 runs or HP joined from >2 HPs
    hp_base = runs1[p_run_ind - 1]['value']
    helper = lambda x: runs1[x]['value'] in ['-', hp_base]
    run_inds_in_hp = [i for i in itertools.takewhile(helper, range(p_run_ind - 1, 0, -1))]
    run_inds_in_hp += [i for i in itertools.takewhile(helper, range(p_run_ind, len(runs1)))]
    runs1_in_hp = runs1[run_inds_in_hp]
    runs1_in_hp = runs1_in_hp[np.where(runs1_in_hp['value'] == hp_base)]
    hp_len = np.sum(runs1_in_hp['length'])
    hp_start_ind = runs1_in_hp[0]['start']
    hp_end_ind = runs1_in_hp[-1]['start'] + runs1_in_hp[-1]['length'] - 1

    # check that the HPs in the query and ref actually differ (some insertions
    # might preserve the HP length e.g. CG--GC -> CGGCGC, in which case the HP
    # is not a split/join
    runs2_in_hp = rle(seq2[hp_start_ind: hp_end_ind + 1])
    runs2_in_hp = runs2_in_hp[np.where(runs2_in_hp['length'] > 1)]
    if (len(runs2_in_hp) > 0 and
        np.any(np.logical_and(runs2_in_hp['length'] == hp_len,
                              runs2_in_hp['value'] == hp_base))):

        return hp_kls

    # we have a split/join if any indels are within hp
    if any([i > hp_start_ind and i < hp_end_ind for i in errors[key]]):
        # if any indels are outside HP run, we have a mess
        if min(errors[key]) < hp_start_ind or max(errors[key]) > hp_end_ind:
            subtype = 'messy'
        elif are_adjacent(errors[key]):
            if len(errors[key]) == 1:
                subtype = 'simple'
            else:
                subtype = 'multi'
        else:
            subtype = 'complex'
        hp_kls = '{} {} HP {} ({}{})'.format(subtype, key, _type_[key], hp_base, hp_len)

    return hp_kls


def get_match_line_and_err_index(context):
    match_line = ''
    errors = defaultdict(list)
    k = ''
    for i, (q, r) in enumerate(zip(context.qb, context.rb)):
        if q == r:
            match_line += _match_
        elif q == '-' or r == '-':  # indel
            match_line += _indel_
            k = 'del' if q == '-' else 'ins'
        else:  # sub
            match_line += _sub_
            k = 'sub'
        if match_line[-1] != _match_:
            errors[k].append(i)
        if i == context.p_i:  # this is the central error we are classifying
            p_k = k
    return match_line, errors, p_k


def preprocess_error(p, aln, search_by_q, offset=10):
    """
    :param p: int, position (ref position, or query position)
    :param aln: iterable of `AlignPos` objects.
    :param search_by_q: bool, whether to search by query position (typically done if qi is None).
    :returns: `Context` object
    """
    s, e, p_i = _get_context_bounds(p, aln, search_by_q, offset)
    sl = aln[s:e]
    qi, qb, ri, rb = zip(*sl)

    return Context(p_i, qb, rb)


def simple_klass(adjacent, n, err_type, indel_sizes):
    if not adjacent:
        descr = 'complex'
    else:
        descr = 'simple' if n == 1 else 'multi'
    if 'ins' in err_type or 'del' in err_type:
        size = _get_size(n, indel_sizes)
        klass = "{} {} {}".format(descr, err_type, size)
    else:
        klass = "{} {}".format(descr, err_type)

    return klass


def classify_error(context, indel_sizes=None):
    """Classify error within an alignment.

    :param context: `Context` object
    :indel_sizes: iterable of int, for binning indel sizes.
        indels >= to indel_sizes[0] will not be considered as HP splitting/joining indels
    :returns: (str reference_context,
               str match_line,
               str query_context,
               dict counts of sub/ins/del within context)
    """
    if indel_sizes is None:
        indel_sizes = _indel_sizes_

    p_i, qb, rb = context
    match_line, errors, p_k = get_match_line_and_err_index(context)

    # find homopolymers within context
    qb_runs = rle([b.upper() for b in qb])
    rb_runs = rle([b.upper() for b in rb])
    qb_hps = qb_runs[np.where(qb_runs['length'] > 1)]
    rb_hps = rb_runs[np.where(rb_runs['length'] > 1)]
    cols = ['value', 'length']
    hp_changed = len(qb_hps) != len(rb_hps) or np.any(qb_hps[cols] != rb_hps[cols])
    # hp might also be changed if we have an ins in the middle of a ref hp
    if not hp_changed:
        hp_changed = any([(rb_runs[i]['value'] == '-' and
                           rb_runs[i - 1]['value'] == rb_runs[i + 1]['value'])
                          for i in range(1, len(rb_runs) - 1)])
    # hp might also be changed if we have an del in the middle of a query hp
    if not hp_changed:
        hp_changed = any([(qb_runs[i]['value'] == '-' and
                           qb_runs[i - 1]['value'] == qb_runs[i + 1]['value'])
                          for i in range(1, len(qb_runs) - 1)])

    hp_kls = None

    # Try to class the error, make it up as we go along!
    dels = len(errors['del'])
    ins = len(errors['ins'])
    indels = dels + ins
    subs = len(errors['sub'])
    # is this position part of a hp in query or ref?
    qp_is_hp = is_in_hp(qb, p_i)
    rp_is_hp = is_in_hp(rb, p_i)
    hp_subtypes = {(False, False): '',
                   (True, False): 'RHP',
                   (False, True): 'QHP',
                   (True, True): 'HP',
    }

    hp_subtype = hp_subtypes[(rp_is_hp, qp_is_hp)]
    err_type = '{} {}'.format(hp_subtype, p_k) if hp_subtype != '' else p_k
    kls = 'unknown ({})'.format(err_type)

    if (subs > 0 and indels > 0) or (ins > 0 and dels >0):
        kls = 'complex mess ({})'.format(err_type)

    elif indels == 0:  # we have a sub
        adjacent = are_adjacent(errors['sub'])
        kls = simple_klass(adjacent, subs, err_type, indel_sizes)
        if hp_changed:
            hp_kls = classify_hp_sub(p_i, adjacent, errors, match_line, rb_runs,
                                     qb_runs, qp_is_hp, rp_is_hp)
            kls = hp_kls if hp_kls is not None else kls

    elif dels == 0:
        adjacent = are_adjacent(errors['ins'])
        if hp_changed and ins < indel_sizes[0]:
            hp_kls = classify_hp_indel(p_i, 'ins', errors, rb_runs, qb)
            kls = hp_kls if hp_kls is not None else kls
        if hp_kls is None:
            kls = simple_klass(adjacent, ins, err_type, indel_sizes)
            if adjacent:
                i = errors['ins'][0]
                b = qb[i]
                q_hp_len = len([x for x in itertools.takewhile(lambda x: x==b, qb[i:])])
                r_hp_len = len([x for x in itertools.takewhile(lambda x: x==b, rb[i+ins:])])
                inserted = qb[i:i + ins]
                # get ref_after insertions to check if we have a repeat
                repeat_start = i + ins
                repeat_end = repeat_start + ins
                if repeat_end <= len(rb):
                    ref_after = rb[repeat_start: repeat_end]
                else:  # we don't have enough context after the insertion
                    ref_after = []

                if (q_hp_len > 1 or r_hp_len > 1) and all([qb[j] == b and rb[j] == '-' for j in errors['ins']]):
                    kls = "HP ins ({}{}->{})".format(b, r_hp_len, q_hp_len)
                elif inserted == ref_after:
                    kls = "fwd repeat ins len {}".format(ins)
                elif inserted == ref_after[::-1]:
                    kls = "rev repeat ins len {}".format(ins)

    elif ins == 0:
        adjacent = are_adjacent(errors['del'])
        if hp_changed and dels < indel_sizes[0]:
            hp_kls = classify_hp_indel(p_i, 'del', errors, qb_runs, rb)
            kls = hp_kls if hp_kls is not None else kls
        if hp_kls is None:
            kls = simple_klass(adjacent, dels, err_type, indel_sizes)
            if adjacent:
                i = errors['del'][0]
                b = rb[i]
                # we use aln_rb instead of rb, so we can find the length
                # of HPs which extent beyond the context.
                hp_len = len([x for x in itertools.takewhile(lambda x: x==b, rb[i:])])
                deleted = rb[i:i + dels]
                # get ref_after deletions to check if we have a repeat
                repeat_start = i + dels
                repeat_end = repeat_start + dels
                if repeat_end <= len(rb):
                    ref_after = rb[repeat_start: repeat_end]
                else:  # we don't have enough context after the insertion
                    ref_after = []
                if hp_len > 1:
                    if all([rb[j] == b for j in errors['del']]):
                        kls = "HP del ({}{}->{})".format(b, hp_len, hp_len - dels)
                elif deleted == ref_after:
                    kls = "fwd repeat del len {}".format(dels)
                elif deleted == ref_after[::-1]:
                    kls = "rev repeat del len {}".format(dels)

    return ''.join(rb), match_line, ''.join(qb), errors, kls


def _get_size(n, sizes):
    l_i = np.searchsorted(sizes, n)
    if l_i == len(sizes):
        size = "> {}".format(sizes[-1])
    else:
        size = "<= {}".format(sizes[l_i])
    return size


def _process_read(bam, outdir, read_range, bed_file=None):
    """Load an alignment from bam and return result of `_process_seg`.

    :param bam: str, bam file.
    :param outdir: str, output path
    :param read_range: (int, int, int), process number, range of alignments to process.
    :param bed_file: path to .bed file of regions to include in analysis.
    :returns: result of `_process_seg`.
    """
    trees = None
    if bed_file is not None:
        trees = intervaltrees_from_bed(bed_file)

    total_ref_length = defaultdict(int)
    total_n_ref_sites_masked = defaultdict(int)
    error_count = defaultdict(Counter)

    db_path = os.path.join(outdir, 'error_catalogue_db_{}.txt'.format(read_range[0]))
    txt_path = os.path.join(outdir, 'error_catalogue_{}.txt'.format(read_range[0]))
    db_fh = open(db_path, 'w')
    txt_fh = open(txt_path, 'w')
    headers = get_headers()

    with pysam.AlignmentFile(bam, 'rb') as bam_obj:
        for i, rec in enumerate(bam_obj.fetch(until_eof=True)):
            if i < read_range[1]:
                continue
            elif i == read_range[2]:
                break

            if rec.is_unmapped or rec.is_supplementary or rec.is_secondary:
                return

            if bed_file is not None:
                if not rec.reference_name in trees or not trees[rec.reference_name].has_overlap(rec.reference_start, rec.reference_end):
                    #sys.stderr.write('read {} does not overlap with any regions in bedfile\n'.format(rec.query_name))
                    continue
                tree = trees[rec.reference_name]
            else:
                tree = None

            seg = AlignSeg(rname=rec.reference_name, qname=rec.query_name,
                           pairs=list(get_trimmed_pairs(rec)), rlen=rec.reference_length
                          )
            logging.debug('Loaded query {}'.format(seg.qname))

            seg_result = _process_seg(seg, db_fh, txt_fh, headers, tree)
            if seg_result is None:
                continue

            ref_name, ref_length, counts, n_masked = seg_result
            
            error_count[ref_name].update(counts)
            total_ref_length[ref_name] += ref_length
            total_n_ref_sites_masked[ref_name] += n_masked

    db_fh.close()
    txt_fh.close()

    return error_count, total_ref_length, total_n_ref_sites_masked


def _process_seg(seg, db_fh, txt_fh, headers, tree=None):
    """Classify and count errors within an `AlignSeg` object.

    :param seg: `AlignSeg` object.
    :param db_fh: file object, catalogue db file
    :param txt_fh: file object, catalogue file
    :param headers: list of tuples, error properties and their label ids
    :param tree: `ncls.NCLS` object of regions to analyse.
    :returns: (seg.rname, aligned_ref_len, error_count, n_masked)
        error_count: `Counter` of error classes
        n_masked: number of reference positions excluded by tree.
    """
    error_count = Counter()
    err_result = get_errors(seg.pairs, tree)
    if err_result is None:  # no matches within bed regions
        logging.debug('Skipping {} since all bed regions were deleted'.format(
                      seg.qname))
        return None

    pos_and_errors, aligned_ref_len, n_masked = err_result
    for ri, qi, error, approx_pos in pos_and_errors:
        ref, match, read, counts, klass = classify_error(preprocess_error(
            ri if ri is not None else qi, seg.pairs, search_by_q=(ri is None)
        ))

        rp = ri
        qp = qi
        if rp is None:
            rp = "~{}".format(approx_pos[0])
        if qp is None:
            qp = "~{}".format(approx_pos[1])

        e = Error(rp=rp, rname=seg.rname, qp=qp, qname=seg.qname, ref=ref,
                  match=match, read=read, counts=counts, klass=klass,
                  aggr_klass=get_aggr_klass(klass))
        error_count[klass] += 1

        db_fh.write(_sep_.join((str(h[1](e)) for h in headers)) + '\n')
        txt_fh.write("Ref Pos: {}, {} Pos {}, {}, {}\n".format(e.rp, e.qname, e.qp, e.klass, e.aggr_klass))
        txt_fh.write(e.ref + "\n")
        txt_fh.write(e.match + "\n")
        txt_fh.write(e.read + "\n")
        txt_fh.write(".\n")

    if tree is None:
        assert seg.rlen == aligned_ref_len

    logging.debug('Done processing {} aligned to {}'.format(seg.qname, seg.rname))
    return seg.rname, aligned_ref_len, error_count, n_masked


def qscore(d):
    """Calculate a qscore"""
    with warnings.catch_warnings():
        # some values might be zero
        warnings.simplefilter("ignore")
        q = -10 * np.log10(d)
    return q


def analyze_counts(counts, total_ref_length):
    df = pd.DataFrame({'klass': list(counts.keys()),
                       'count': list(counts.values())})
    df['err_rate'] = df['count'] / total_ref_length
    df['fraction_of_total'] = df['count'] / df['count'].sum()
    df.sort_values('count', ascending=True, inplace=True)
    df['remaining_count'] = df['count'].cumsum()
    df['remaining_err_rate'] = df['err_rate'].cumsum()
    df['err_rate_q'] = qscore(df['err_rate'])
    df['remaining_err_rate_q'] = qscore(df['remaining_err_rate'])
    df.sort_values('remaining_count', ascending=False, inplace=True)
    return df


def plot_summary(df, outdir, prefix, ref_len):
    """Create a plot showing Q-scores as largest remaining
    error klass is removed"""

    fig, ax = plt.subplots()
    fig.subplots_adjust(left=0.3)
    y_pos = np.arange(len(df) + 1)
    no_error_score = -10 * np.log10(1/ref_len)
    ax.barh(y_pos, np.concatenate([df['remaining_err_rate_q'].to_numpy(), [no_error_score]]), align='center', color='green', ecolor='black')
    ax.set_xlabel('Q(Accuracy)')
    ax.set_ylabel('Error Class')
    ax.set_ylim((y_pos[0]-0.5, y_pos[-1]+0.5))
    ax.set_yticks(y_pos)
    ax.set_yticklabels(['total error'] + list(df['klass']))
    ax.invert_yaxis()  # labels read top-to-bottom
    xstart, xend = ax.get_xlim()
    ystart, yend = ax.get_ylim()
    ax.text(xend - 2.25, ystart - 0.25, '+')
    ax.set_title('Q-score after removing error class')
    fp = os.path.join(outdir, '{}_remaining_errors.png'.format(prefix))
    fig.savefig(fp)
    plt.close()


def get_aggr_counts(total_counts):
    # get errors per type
    aggregate_counts = Counter()
    max_indel = max(_indel_sizes_)
    for key, val in total_counts.items():
        aggregate_counts[get_aggr_klass(key)] += val
    return aggregate_counts


def get_aggr_klass(klass):
    max_indel = max(_indel_sizes_)
    for error_type, is_type in _error_groups_:
        if is_type(klass):
            if '>' in error_type or '<' in error_type:
                aggr_klass = '{} {}'.format(error_type, max_indel)
            else:
                aggr_klass = error_type
            break
    return aggr_klass


def analyse_errors(args):
    # load and merge existing counts
    error_count = defaultdict(Counter)
    total_ref_length = Counter()
    total_n_ref_sites_masked = Counter()
    for pkl in args.pkl:
        with open(pkl, 'rb') as fh:
            counts = pickle.load(fh)
        
        # pkl format:
        # {'ref_lengths': total_ref_length,
        #  'n_ref_sites_masked': total_n_ref_sites_masked,
        #  'counts': {'by_ref': error_count,
        #             'by_ref_aggr': aggr_by_ref,
        #             'total': total_counts,
        #             'total_aggr': aggregate_counts,
        #            }
        # }
        for ref, d_counts in counts['counts']['by_ref'].items():
            error_count[ref].update(d_counts)
        total_ref_length.update(counts['ref_lengths'])
        total_n_ref_sites_masked.update(counts['n_ref_sites_masked'])

    aggr_and_output(args, error_count, total_ref_length, total_n_ref_sites_masked)


def get_headers():
    # make an approximate position into int
    helper = lambda x: int(x.replace('~','')) if isinstance(x, str) else x

    return [('ref_name', attrgetter('rname')),
            ('ref_pos', helper(attrgetter('rp'))),
            ('ref_context', attrgetter('ref')),
            ('query_name', attrgetter('qname')),
            ('query_pos', helper(attrgetter('qp'))),
            ('query_context', attrgetter('read')),
            ('class', attrgetter('klass')),
            ('aggr_class', attrgetter('aggr_klass')),
            ('n_ins', lambda e: len(e.counts['ins'])),
            ('n_del', lambda e: len(e.counts['del'])),
            ('n_sub', lambda e: len(e.counts['sub'])),
            ('context_len', lambda e: len(e.ref)),
    ]


def count_errors(args):
    # create a slice of reads to process in each thread to avoid looping through
    # bam n read times and reduce mp overhead
    ranges = [(0, 0, float('inf'))]
    if args.threads > 1:
        with pysam.AlignmentFile(args.bam) as bam:
            n_reads = bam.count(until_eof=True)
        n_reads_per_proc = ceil(n_reads / args.threads)
        ranges = [(thread, thread * n_reads_per_proc, min((thread + 1) * n_reads_per_proc, n_reads))
                   for thread in range(args.threads)]

    total_ref_length = defaultdict(int)
    total_n_ref_sites_masked = defaultdict(int)
    error_count = defaultdict(Counter)

    # record draft start position of each long multi indel
    multi_errs = {}

    f = partial(_process_read, args.bam, args.outdir, bed_file=args.bed)

    with concurrent.futures.ProcessPoolExecutor(max_workers=args.threads) as ex:
        for returned in ex.map(f, ranges):
            if returned is None:
                continue
            else:
                counts, ref_length, n_masked = returned

            for ref_name in counts:
                error_count[ref_name].update(counts[ref_name])
                total_ref_length[ref_name] += ref_length[ref_name]
                total_n_ref_sites_masked[ref_name] += n_masked[ref_name]

    db_fh = open(os.path.join(args.outdir, 'error_catalogue_db.txt'), 'w')
    db_fh.write(_sep_.join((h[0] for h in get_headers())) + '\n')
    merge_catalogues(db_fh, [os.path.join(args.outdir, 'error_catalogue_db_{}.txt'.format(n)) for n in range(args.threads)])
    db_fh.close()

    txt_fh = open(os.path.join(args.outdir, 'error_catalogue.txt'), 'w')
    merge_catalogues(txt_fh, [os.path.join(args.outdir, 'error_catalogue_{}.txt'.format(n)) for n in range(args.threads)])
    txt_fh.close()

    aggr_and_output(args, error_count, total_ref_length, total_n_ref_sites_masked)


def merge_catalogues(outfile, filenames):
    for f in filenames:
        try:
            with open(f, 'r') as infile:
                shutil.copyfileobj(infile, outfile)
        except:
            logging.info("Error merging file {}".format(f))
        else:
            logging.info("Merged file {}, deleting".format(f))
            os.remove(f)


def aggr_and_output(args, error_count, total_ref_length, total_n_ref_sites_masked):
    total_counts = Counter()
    aggr_by_ref = {}
    for ref_name, counts in error_count.items():
        df = analyze_counts(counts, total_ref_length[ref_name])
        fp = os.path.join(args.outdir, '{}_error_summary.txt'.format(ref_name))
        df.to_csv(fp, sep=_sep_, index=False)

        aggr_by_ref[ref_name] = get_aggr_counts(counts)
        df = analyze_counts(aggr_by_ref[ref_name], total_ref_length[ref_name])
        fp = os.path.join(args.outdir, '{}_aggr_error_summary.txt'.format(ref_name))
        df.to_csv(fp, sep=_sep_, index=False)
        plot_summary(df, args.outdir, '{}_aggr'.format(ref_name),
                     ref_len=total_ref_length[ref_name])

        total_counts.update(counts)

    df = analyze_counts(total_counts, sum(total_ref_length.values()))
    fp = os.path.join(args.outdir, '{}_error_summary.txt'.format('total'))
    df.to_csv(fp, sep=_sep_, index=False)

    aggregate_counts = get_aggr_counts(total_counts)
    df = analyze_counts(aggregate_counts, sum(total_ref_length.values()))
    fp = os.path.join(args.outdir, '{}_aggr_error_summary.txt'.format('total'))
    df.to_csv(fp, sep=_sep_, index=False)
    plot_summary(df, args.outdir, '{}_aggr'.format('total'),
                 ref_len=sum(total_ref_length.values()))

    # save counts to yaml for any further analysis
    to_save = {'ref_lengths': total_ref_length,
               'n_ref_sites_masked': total_n_ref_sites_masked,
               'counts': {'by_ref': error_count,
                          'by_ref_aggr': aggr_by_ref,
                          'total': total_counts,
                          'total_aggr': aggregate_counts,
                         }
               }
    with open(os.path.join(args.outdir, 'counts.pkl'), 'wb') as fh:
         pickle.dump(to_save, fh)


def main():
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    parser = argparse.ArgumentParser(
        prog='catalogue_errors',
        description='Create a catalogue of all query errors in a bam.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command')
    subparsers.required = True
    
    cparser = subparsers.add_parser('count',
        help='Count query errors in a bam. ',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    cparser.set_defaults(func=count_errors)
    cparser.add_argument('bam', help='Input alignments (aligned to ref).')
    cparser.add_argument('--bed', default=None, help='.bed file of reference regions to include.')
    cparser.add_argument('-t', '--threads', type=int, default=1, help='Number of threads for parallel execution.')
    cparser.add_argument('-o', '--outdir', default='error_catalogue', help='Output directory.')

    aparser = subparsers.add_parser('analyse',
        help='Analyse existing counts, optionally merging multiple counters.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    aparser.set_defaults(func=analyse_errors)
    aparser.add_argument('pkl', nargs='+', help='Input .pkl file(s).')
    aparser.add_argument('-o', '--outdir', default='error_catalogue',
        help="Output directory (will be created).")

    args = parser.parse_args()
    os.mkdir(args.outdir)
    args.func(args)
    logging.info('All done, check {} for output.'.format(args.outdir))


if __name__ == '__main__':
    main()


class ClassifyErrorTest(unittest.TestCase):

    def setUp(self):
        pass


    def test_hp_del_6_5(self):
        rb = 'ACAACAGCAGAAAAAACAGGA'
        qb = 'ACAACAGCAG-AAAAACAGGA'
        p_i = 10
        expected = 'HP del (A6->5)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_hp_del_2_0(self):
        rb = 'CACTTTCGGCTTGAGGATCA'
        qb = 'CACTTTCGGC--GAGGATCA'
        p_i = 10
        expected = 'HP del (T2->0)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_multi_ins(self):
        rb = 'ATGTAATGCC---AAGCTTA'
        qb = 'ATGTAATGCCAGAAAGCTT'
        p_i = 10
        expected = 'multi ins <= 5'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_multi_del(self):
        rb = 'ATGTAATGCCAGAAAGCTT'
        qb = 'ATGTAATGCC---AAGCTTA'
        p_i = 10
        expected = 'multi del <= 5'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_hp_ins(self):
        rb = 'CACCTGGTGC-AAAAGAGAG'
        qb = 'CACCTGGTGCAAAAAGAGAG'
        p_i = 10
        expected = 'HP ins (A4->5)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_sub_split(self):
        rb = 'TAATCTGGCCcCTGCAATGC'
        qb = 'TAATCTGGCCTCTGCAATGC'
        p_i = 10
        expected = 'sub HP split (C4)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_sub_swap_trc(self):
        rb = 'ACTGCGTACCtTTTGTATAAT'
        qb = 'ACTGCGTACCCTTTGTATAAT'
        p_i = 10
        expected = 'HP sub swp (T4->C2,1)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_sub_swap_trc2(self):
        rb = 'ACTGCGTACCttTTGTATAAT'
        qb = 'ACTGCGTACCCCTTGTATAAT'
        p_i = 10
        expected = 'HP sub swp (T4->C2,2)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_sub_swap_ext(self):
        rb = 'ACTGCGTACcTTTTGTATAAT'
        qb = 'ACTGCGTACTTTTTGTATAAT'
        p_i = 9
        expected = 'HP sub swp (T4->C2,1)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_sub_swap_ext(self):
        rb = 'ACTGCGTAccTTTTGTATAAT'
        qb = 'ACTGCGTATTTTTTGTATAAT'
        p_i = 9
        expected = 'HP sub swp (C2->T4,2)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_swap_sub_trc(self):
        rb = 'CGGGCCTTCCCCtTGCCATTCA'
        qb = 'CGGGCCTTCCCCCTGCCATTCA'
        p_i = 12
        expected = 'HP sub swp (T2->C4,1)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_flk_sub_trc(self):
        rb = 'CGAGAAAATCgGGATCGTTG'
        qb = 'CGAGAAAATCAGGATCGTTG'
        p_i = 10
        expected = 'flk HP sub trc (G3->2)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_flk_sub_ext(self):
        rb = 'ATGCAACAAGcTTACGCTGC'
        qb = 'ATGCAACAAGTTTACGCTGC'
        p_i = 10
        expected = 'flk HP sub ext (T2->3)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_complex_sub(self):
        rb = 'CGAGAAAAAAAGGATCGTTG'
        qb = 'CGAGTATAAAAGGATCGTTG'
        p_i = 4
        expected = 'complex HP sub trc (A7)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_messy_sub(self):
        rb = 'CTAaACtGCcGTG'
        qb = 'CTACACCGCTGTG'
        p_i = 3
        expected = 'messy HP sub trc'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_sub_join(self):
        rb = 'AGGAACGAATcTCTGAAGCG'
        qb = 'AGGAACGAATTTCTGAAGCG'
        p_i = 10
        expected = 'sub HP join (T3)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_not_complete_sub(self):
        rb = 'AGGAACGAATTTCTGAAGCG'
        qb = 'AGGAACGAACCCCTGAAGCG'
        p_i = 10
        expected = 'HP sub swp (T3->C1,3)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_complete_sub(self):
        rb = 'AGGAACGAATTTGTGAAGCG'
        qb = 'AGGAACGAACCCGTGAAGCG'
        p_i = 10
        expected = 'HP sub swp (T3->C0,3)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_messy_sub_ext(self):
        rb = 'CGGGTCTTTTctTTTTCATC'
        qb = 'CGGGTCTTTTTCTTTTCATC'
        p_i = 10
        expected = 'messy HP sub ext'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_del_join(self):
        rb = 'CGGGTCTTTTCTTTTCATC'
        qb = 'CGGGTCTTTT-TTTTCATC'
        p_i = 10
        expected = 'simple del HP join (T8)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_del_join2(self):
        rb = 'CGGGTCTTTTCCTTTTCATC'
        qb = 'CGGGTCTTTT--TTTTCATC'
        p_i = 10
        expected = 'multi del HP join (T8)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_del_join3(self):
        rb = 'CGGGTCTTTTCATTTCATC'
        qb = 'CGGGTCTTTT--TTTCATC'
        p_i = 10
        expected = 'multi del HP join (T7)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_del_join4(self):
        rb = 'CGGGTCTTTTCATTTCATC'
        qb = 'CGG-TCTTTT--TTTCATC'
        p_i = 10
        expected = 'messy del HP join (T7)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_ins_split(self):
        rb = 'CGGGTCTT-TTTCATC'
        qb = 'CGGGTCTTGTTTCATC'
        p_i = 8
        expected = 'simple ins HP split (T5)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_ins_split2(self):
        rb = 'CGGGTCTT--TTTCATC'
        qb = 'CGGGTCTTGGTTTCATC'
        p_i = 8
        expected = 'multi ins HP split (T5)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_ins_split3(self):
        rb = 'CGGGTCTT--TTTCATC'
        qb = 'CGGGTCTTGATTTCATC'
        p_i = 8
        expected = 'multi ins HP split (T5)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_ins_split4(self):
        rb = 'CGGGTCTT--TT-TCATC'
        qb = 'CGGGTCTTGATTGTCATC'
        p_i = 8
        expected = 'complex ins HP split (T5)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_ins_split5(self):
        rb = 'CGGG-TCTT--TT-TCATC'
        qb = 'CGGGGTCTTGATTGTCATC'
        p_i = 9
        expected = 'messy ins HP split (T5)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_fwd_repeat_ins(self):
        rb = 'ACCTATAACG--GCGCGCTG'
        qb = 'ACCTATAACGGCGCGCGCTG'
        p_i = 10
        expected = 'fwd repeat ins len 2'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_fwd_repeat_ins(self):
        rb = 'ACCTATAACG--GCGCGCTG'
        qb = 'ACCTATAACGCGGCGCGCTG'
        p_i = 10
        expected = 'rev repeat ins len 2'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_HP_ins_split6(self):
        rb = 'GCCGATTTTT-TCTCCCGTA'
        qb = 'GCCGATTTTTCTCTCCCGTA'
        p_i = 10
        expected = 'simple ins HP split (T6)'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_complex_HP_del(self):
        rb = 'AGGGGGGGGACTTGAACCCCCACGTC'
        qb = 'A-GGGGGGGACTTGAA-CCCCACGTC'
        p_i = 16
        expected = 'complex RHP del <= 5'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_long_multi_del(self):
        rb = 'ACCCACACACCACACCCACACACCACACCCACACCACACCCACACCACACCCACACACCACACCCACACCACACCCACACACCACACCCACACACCACACCCACACCACACCCACACCACACCCACACACCACACCACACCCACACACCCACACACCACACACTCTCTTACATCTACCTCTACTCTCGCTGTCACACCTTACCCGGCTTTCTGACCGAAATTAAAAAAAATGAAAATGAAATCCTCTTCTTTAGCCCTACAACACTTTTACATAGCCCTAAATAGCCCTAAATAGCCCTCATGTACGTCTCCTCCAAGCCCTATTGACTCTTACCCGGAGTTTCAGCTAAAGGCTATACTTACT'
        qb = 'ACCCACACACCA---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------T'
        p_i = 12
        expected = 'multi del > 100'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)

    def test_long_multi_ins(self):
        rb = 'ACCCACACACCA---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------T'
        qb = 'ACCCACACACCACACCCACACACCACACCCACACCACACCCACACCACACCCACACACCACACCCACACCACACCCACACACCACACCCACACACCACACCCACACCACACCCACACCACACCCACACACCACACCACACCCACACACCCACACACCACACACTCTCTTACATCTACCTCTACTCTCGCTGTCACACCTTACCCGGCTTTCTGACCGAAATTAAAAAAAATGAAAATGAAATCCTCTTCTTTAGCCCTACAACACTTTTACATAGCCCTAAATAGCCCTAAATAGCCCTCATGTACGTCTCCTCCAAGCCCTATTGACTCTTACCCGGAGTTTCAGCTAAAGGCTATACTTACT'
        p_i = 12
        expected = 'multi ins > 100'
        found = classify_error(Context(p_i=p_i, qb=list(qb), rb=list(rb)))[-1]
        self.assertEqual(found, expected)



import argparse
import collections
import concurrent.futures
from functools import partial
from math import ceil
import os
import pickle
import re
import shutil

import matplotlib
matplotlib.use('Agg', force=True)

import  matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np
import pandas as pd
import pysam

from pomoxis import bio
from pomoxis.util import intervaltrees_from_bed
from pomoxis.summary_from_stats import qscore

class AverageScore:
    """Keep track of a simple average of inputs."""

    def __init__(self, cumulative_score=0, count=0):
        self.cumulative_score = cumulative_score
        self.count = count

    def __add__(self, score):
        if isinstance(score, AverageScore):
            self.cumulative_score += score.cumulative_score
            self.count += score.count
        else:
            self.cumulative_score += score
            self.count += 1
        return self

    @property
    def average_score(self):
        """Return the average of the accumulated samples."""
        try:
            return self.cumulative_score / self.count
        except ZeroDivisionError:
            return 0

    def __repr__(self):
        return str(self.average_score)


def get_fraction_correct(counts):
    """Calculate fraction correct vs length from counts."""
    corr_res = []
    for base, c in counts.items():
        if base != 'length':
            for rlen, sub_counts in c.items():
                corr_res.append({
                    'qbase': base,
                    'rlen': rlen,
                    'n': sum(sub_counts.values()),
                    'n_corr': sub_counts[rlen],
                    'fr_corr': sub_counts[rlen] / max(sum(sub_counts.values()), 1)
                    })
    df = pd.DataFrame(corr_res)
    df_fr = df.pivot(index='rlen', columns='qbase', values='fr_corr').reset_index()
    df_n = df.pivot(index='rlen', columns='qbase', values='n').reset_index()
    df_n[np.isnan(df_n)] = 0
    df = pd.merge(df_fr, df_n, on='rlen', suffixes=['', '_n'])
    for b in ['A', 'T', 'G', 'C']:
        if b not in df:
            df.loc[:, b] = np.NaN
            df.loc[:, b + '_n'] = 0
    for b, c in [('A', 'T'), ('G', 'C')]:
        # get total count and weighted-fraction correct for AT and GC
        sum_n_col = b + c + '_n'
        df[sum_n_col] = df[b + '_n'].add(df[c + '_n'], fill_value=0)
        df[b + c] = sum([df[a + '_n'].mul(df[a], fill_value=0) for a in [b, c]]) / df[sum_n_col]
    df['ATGC_n'] = df.loc[:, ('A_n', 'T_n', 'G_n', 'C_n')].sum(axis=1)
    df['ATGC'] = sum([df[b + '_n'].mul(df[b], fill_value=0) for b in ('AT', 'GC')]) / df['ATGC_n']
    return df


def plot_fraction_correct(df, fname, cols=('A', 'T', 'G', 'C')):
    """Plot fraction correct vs length from counts."""
    fig, ax = plt.subplots()
    for col in cols:
        ax.plot(df['rlen'], df[col], 'o-', label=col)
    ax.set_ylabel('Fraction Correct')
    ax.set_xlabel('Homopolymer length')
    ax.legend(frameon=False)
    fig.savefig(fname)
    plt.close(fig)


def get_relative_lengths(counts):
    rows = []
    for q_base, c in counts.items():
        if q_base != 'length':
            for ref_len, qc in c.items():
                data = {'ref_len': ref_len, 'q_base': q_base}
                for query_len, count in qc.items():
                    rows.append(data.copy())
                    rows[-1]['query_len'] = query_len
                    rows[-1]['count'] = count
                    rows[-1]['rel_len'] = query_len - ref_len

    return pd.DataFrame(rows)


def combine_complementary_relative_lengths(data):
    data = data.copy()
    data['is_AT'] = data['q_base'].str.contains(r'[AT]')

    at_gc = data.groupby(['is_AT', 'ref_len', 'rel_len'], as_index=False).agg({"count": "sum"})
    at_gc['q_base'] = ''
    at_gc.loc[np.where(at_gc['is_AT'])[0], 'q_base'] = 'AT'
    at_gc.loc[np.where(np.logical_not(at_gc['is_AT']))[0], 'q_base'] = 'GC'
    del at_gc['is_AT']
    return at_gc


def plot_relative_lengths(data, fname):
    # some panels will be empty for longer HPs so keep track of what lengths and bases are in each row/column
    rows = sorted(data['ref_len'].unique())
    cols = sorted(data['q_base'].unique())
    fig, axes = plt.subplots(
        ncols=len(cols), nrows=len(rows), sharex=True,
        figsize=(4 * len(cols), 2 * len(rows)), squeeze=False)
    for rl, rl_df in data.groupby('ref_len'):
        i = rows.index(rl)
        for qb, qb_df in rl_df.groupby('q_base'):
            j = cols.index(qb)
            ax = axes[i][j]
            ax.bar(qb_df['rel_len'], qb_df['count'])
            ax.set_title('{}{}'.format(qb, rl))
            ax.xaxis.set_tick_params(which='both', labelbottom=True)
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    for ax in axes[-1,:]:
        ax.set_xlabel('Query length relative to reference')
    for ax in axes[:, 0]:
        ax.set_ylabel('Counts')
    fig.tight_layout()
    fig.savefig(fname)
    plt.close(fig)


def get_heatmap(data, data_comp):
    """Transform relative length data into a 2d array of called length vs true length."""
    df = data.pivot(index='query_len', columns=['q_base', 'ref_len'],
        values='count').sort_index(axis=0).sort_index(axis=1)
    column_reindex = pd.MultiIndex.from_product((['A','T','G','C'],
            range(df.columns.levels[1].min(), df.columns.levels[1].max() + 1)))
    df = df.reindex(range(df.index.min(), df.index.max() + 1))
    df[np.isnan(df)] = 0
    df = df.reindex(column_reindex, axis=1)

    df_comp = data_comp.copy()
    df_comp['query_len'] = df_comp['ref_len'] + df_comp['rel_len']
    df_comp = df_comp.pivot(index='query_len', columns=['q_base', 'ref_len'],
        values='count').sort_index(axis=0).sort_index(axis=1)
    column_reindex = pd.MultiIndex.from_product((df_comp.columns.levels[0],
            range(df_comp.columns.levels[1].min(), df_comp.columns.levels[1].max() + 1)),
            names=['q_base', 'ref_len'])
    df_comp = df_comp.reindex(range(df_comp.index.min(), df_comp.index.max() + 1))
    df_comp[np.isnan(df_comp)] = 0
    df_comp = df_comp.reindex(column_reindex, axis=1)

    df_all = df_comp.groupby(level='ref_len', axis=1).sum()
    df_all.columns = pd.MultiIndex.from_product((['ATGC',], df_all.columns), 
        names=['q_base', 'ref_len'])

    df = pd.concat((df, df_comp, df_all), axis=1)
    df[np.isnan(df)] = 0
    return df


def plot_heatmap(df, fname, cols=('A', 'T', 'G', 'C')):
    """Plot heatmaps of called length vs true length."""
    fig, axes = plt.subplots(1, len(cols), figsize=(6 * len(cols), 4),
            squeeze=False)
    for col, ax in zip(cols, axes.flatten()):
        if col in df.columns.levels[0]:
            extent = [df[col].columns.min() - 0.5, df[col].columns.max() + 0.5,
                      df[col].index.min() - 0.5, df[col].index.max() + 0.5]
            im = ax.imshow(df[col] / df[col].sum(), origin='lower',
                           extent=extent, vmin=0, vmax=1)
        ax.set_ylabel('Called Length')
        ax.set_xlabel('True Length')
        plt.colorbar(im, ax=ax, label='probability')
        ax.title.set_text(col)
    fig.tight_layout()
    fig.savefig(fname)
    plt.close(fig)


def get_errors_by_length(data, aligned_len):
    """Calculate HP errors and Q scores separately for each length."""
    e = data.copy()
    e['total_err'] = e['count'] * np.abs(e['rel_len'])
    e = e[['ref_len', 'count', 'total_err']].groupby('ref_len').sum().reset_index()
    e['total_err_rate'] = e['total_err'] / aligned_len
    e['total_err_to'] = np.cumsum(e['total_err'])
    e['total_err_from'] = np.cumsum(e['total_err'].iloc[::-1])
    e['total_err_rate_to'] = e['total_err_to'] / aligned_len
    e['total_err_rate_from'] = e['total_err_from'] / aligned_len
    e['QHP'] = qscore(e['total_err_rate'])
    e['QHP_to'] = qscore(e['total_err_rate_to'])
    e['QHP_from'] = qscore(e['total_err_rate_from'])
    e['QHPmax'] = qscore(1. / aligned_len)
    return e


def plot_errors_by_length(e, fname):
    """Plot Q score associated with HP errors as a function on HP length."""
    fig, ax = plt.subplots()
    e.plot('ref_len', ['QHP', 'QHP_to', 'QHP_from', 'QHPmax'],
            ax=ax, style=['-o', '-o', '-o', ':k'],
            label=['QHP(n)', 'QHP(<=n)', 'QHP(>=n)', 'acc. lim.'],
            ylabel="Q score",
            xlabel="Homopolymer length")
    fig.savefig(fname)
    plt.close(fig)


def count_homopolymers(args):
    os.mkdir(args.output_dir)

    # create a slice of reads to process in each thread to avoid looping through
    # bam n read times and reduce mp overhead
    ranges = [(0, 0, float('inf'))]
    if args.threads > 1:
        with pysam.AlignmentFile(args.bam) as bam:
            n_reads = bam.count(until_eof=True)
        n_reads_per_proc = ceil(n_reads / args.threads)
        ranges = [(thread, thread * n_reads_per_proc, min((thread + 1) * n_reads_per_proc, n_reads))
                   for thread in range(args.threads)]

    headers = ['query_name', 'ref_name', 'ref_start', 'rev_comp', 'query_start',
               'ref_base', 'query_base', 'ref_len', 'query_len']

    prefix = os.path.join(args.output_dir, 'hp')

    total_score = AverageScore()
    total_counts = collections.defaultdict(counts_factory)
    total_aligned_length = 0

    f = partial(process_bam, args.bam, prefix, args.homo_len, bed_file=args.bed)

    with concurrent.futures.ProcessPoolExecutor(max_workers=args.threads) as ex:
        for returned in ex.map(f, ranges):
            if returned is None:
                continue
            else:
                score, counts, aligned_length = returned
                for base in counts:
                    for rlen in counts[base]:
                        total_counts[base][rlen].update(counts[base][rlen])
                total_score += score
                total_aligned_length += aligned_length

    print("Found {} homopolymers in {}. Average score: {}".format(
        total_score.count, args.bam, total_score))
    total_counts['length'] = total_aligned_length

    with open(prefix + '_catalogue.txt', 'w') as txt_fh:
        txt_fh.write('\t'.join(headers) + "\n")
        for f in [prefix + '_catalogue_{}.txt'.format(n) for n in range(args.threads)]:
            try:
                with open(f, 'r') as infile:
                    shutil.copyfileobj(infile, txt_fh)
            except:
                print("Error merging file {}".format(f))
            else:
                print("Merged file {}, deleting".format(f))
                os.remove(f)

    save_counts(total_counts, prefix + '_counts.pkl')
    analyse_counts(total_counts, prefix, total_aligned_length)


def save_counts(counts, fp):
    # convert counts from nested defaultdict with factory function to plain dict
    counts = {k: v if type(v) == int else dict(v) for k,v in counts.items()}
    with open(fp, 'wb') as pkl_fh:
        pickle.dump(counts, pkl_fh)


def load_counts(pkl):
    with open(pkl, 'rb') as fh:
        counts = pickle.load(fh)
    return counts


def merge_counts(pkl_files):
    # load and merge existing counts
    counts = collections.defaultdict(counts_factory)
    for pkl in pkl_files:
        c = load_counts(pkl)
        for base, base_counts in c.items():
            if base == 'length':
                if 'length' not in counts:
                    counts['length'] = base_counts
                else:
                    counts['length'] += base_counts
            else:
                for rlen, rlen_counts in base_counts.items():
                    counts[base][rlen].update(rlen_counts)
    return counts


def analyse_homopolymers(args):
    os.mkdir(args.output_dir)
    prefix = os.path.join(args.output_dir, 'hp')
    counts = merge_counts(args.pkl)
    save_counts(counts, prefix + '_counts.pkl')
    if 'length' in counts:
        aligned_length = counts['length']
    else:
        aligned_length = None
    analyse_counts(counts, prefix, aligned_length)


def analyse_counts(counts, prefix, length=None):
    fr_correct = get_fraction_correct(counts)
    fr_correct.to_csv(prefix + '_correct_vs_len.txt', sep='\t', index=False)
    plot_fraction_correct(fr_correct, prefix + '_correct_vs_len.png')
    plot_fraction_correct(fr_correct, prefix + '_correct_vs_len_comp_pairs.png', cols=('AT', 'GC'))
    plot_fraction_correct(fr_correct, prefix + '_correct_vs_len_all.png', cols=('ATGC',))

    data = get_relative_lengths(counts)
    data.to_csv(prefix + '_rel_len_counts.txt', sep='\t', index=False)
    plot_relative_lengths(data, prefix + '_rel_len_counts.png')

    data_comp = combine_complementary_relative_lengths(data)
    data_comp.to_csv(prefix + '_rel_len_counts_comp_pairs.txt', sep='\t', index=False)
    plot_relative_lengths(data_comp, prefix + '_rel_len_counts_comp_pairs.png')

    heatmap = get_heatmap(data, data_comp)
    heatmap.loc[:, ['A', 'T', 'G', 'C']].to_csv(prefix + '_heatmap.txt', sep='\t')
    heatmap.loc[:, ['AT', 'GC']].to_csv(prefix + '_heatmap_comp_pairs.txt', sep='\t')
    heatmap.loc[:, 'ATGC'].to_csv(prefix + '_heatmap_all.txt', sep='\t')
    plot_heatmap(heatmap, prefix + '_call_distribution.png')
    plot_heatmap(heatmap, prefix + '_call_distribution_comp_pairs.png', cols=('AT', 'GC'))
    plot_heatmap(heatmap, prefix + '_call_distribution_all.png', cols=('ATGC',))

    if length is not None:
        err_by_length = get_errors_by_length(data, length)
        err_by_length.to_csv(prefix + '_qhp_vs_len.txt', sep='\t', index=False)
        plot_errors_by_length(err_by_length, prefix + '_qhp_vs_len.png')


def get_next_aligned_base(align, seq_start):
    next_base = seq_start
    try:
        while align[next_base][0] is None:
            next_base += 1
    except IndexError:
        # We don't have any more alignments in the read!
        next_base = max(align, key=lambda x: x[0])
    return align[next_base][0]


def find_homopolymers(seq, min_length, alphabet='ACGT'):
    """
    :param seq: the sequence to search
    :param min_length: minimum hp length to find
    :param alphabet: alphabet of bases to look for

    :returns:
    """
    # backreference to force separation between different letter patterns
    # e.g. 'GTAAAAACCCCCTG' -> 'AAAAA' 'CCCCC' not 'AAAAACCCCC'
    homopolymer_expr = r"([%s])\1{%s,}" % (alphabet, min_length - 1)
    return re.finditer(homopolymer_expr, seq)


def get_longest_homopolymer(seq, base):
    try:
        length, offset = max(
            [(len(m.group()), m.start())
                for m in find_homopolymers(seq, 1, alphabet=base)],
            key=lambda x: x[0])
    except ValueError:
        # No matches found
        length, offset = (0, 0)
    return length, offset


def get_query_coords(align_pairs, seq_start, hp_len, base):
    # Find the bases around the seq_start which align to the reference and match the base
    alignment_region = align_pairs[seq_start - 1: seq_start + hp_len + 1]
    alignment_matches = [x[0] for x in alignment_region if x[0] is not None and x[2] == base]
    qstart = min(alignment_matches)
    qend = max(alignment_matches)
    return qstart, qend


def get_query_region(query_sequence, query_start, query_end, base):
    extend_fwd = 1
    extend_backward = -1
    while (query_end + extend_fwd + 1) < len(query_sequence) and \
            query_sequence[query_end + extend_fwd] == base:
        extend_fwd += 1
    while query_sequence[query_start + extend_backward] == base and \
            (query_start + extend_backward) > 0:
        extend_backward -= 1
    return query_sequence[query_start + extend_backward + 1:query_end + extend_fwd]


def counts_factory():
    return collections.defaultdict(collections.Counter)


def process_bam(bam, prefix, homo_len, read_range, bed_file=None):
    """Count HPs in a chunk of reads.

    :param bam: str, path to bam file
    :param prefix: str, prefix for output paths
    :param homo_len: int, minimum ref length of HPs to count
    :param read_range: (int, int, int), chunk number, read index start and end
    :param bed_file: path to .bed file, regions to analyse
    :returns: score, counts, aligned_ref_len
        score, AverageScore object
        counts, defaultdict of defaultdicts of Counters, counts by contig, ref length,
            query length
        aligned_ref_len, int
    """
    counts = collections.defaultdict(counts_factory)
    score = AverageScore()
    aligned_ref_len = 0

    catalogue_path = prefix + '_catalogue_{}.txt'.format(read_range[0])
    out_fh = open(catalogue_path, 'w')

    trees = None
    if bed_file is not None:
        trees = intervaltrees_from_bed(bed_file)

    with pysam.AlignmentFile(bam, 'r') as bam_obj:
        aligned_ref_len = 0
        for i, seq in enumerate(bam_obj.fetch(until_eof=True)):
            if i < read_range[1]:
                continue
            elif i == read_range[2]:
                break

            if seq.is_secondary or seq.is_supplementary or seq.is_unmapped:
                continue
            seq_reference_start = seq.reference_start  # avoid recomputing for each HP
            seq_reference_end = seq.reference_end  # avoid recomputing for each HP

            if trees is not None and (seq.reference_name not in trees or not
                trees[seq.reference_name].has_overlap(seq_reference_start, seq_reference_end)):
                continue

            refseq = seq.get_reference_sequence().upper()
            align_pairs = seq.get_aligned_pairs(with_seq=True)
            # lookup is a mapping of the refseq indices to the query sequence indicies
            lookup = np.array(
                [i for i, v in enumerate(align_pairs) if v[1] is not None])

            # find all the homopolymers in the reference and then attempt to match
            # each one in the query sequence
            for homo_ref in find_homopolymers(refseq, min_length=homo_len):
                base = homo_ref.group()[0]
                ref_hp_len = len(homo_ref.group())
                ref_start = homo_ref.start()
                align_start = lookup[ref_start]

                try:
                    query_start, query_end = get_query_coords(
                        align_pairs, align_start, ref_hp_len, base)
                    query_region = get_query_region(
                        seq.query_sequence, query_start, query_end, base)

                    # We find the longest hp in the region to avoid insertions
                    # breaking our results
                    # e.g. ACAAAAA -> (5mer, offset=2) not (1mer offset=0)
                    # nor 7mer!
                    call_len, offset = get_longest_homopolymer(query_region, base)
                    query_start += offset
                except ValueError:
                    # If we are at the end of the alignment, skip this hp
                    if align_start == 0 or align_start + ref_hp_len == len(align_pairs):
                        continue
                    # If we don't find any matching base then we return the
                    # next query position that aligns
                    query_start = get_next_aligned_base(align_pairs, align_start)
                    call_len = 0
                orginal_base = base
                # Output the results
                if seq.is_reverse:
                    query_start = seq.query_length - query_start - call_len
                    base = bio.comp[base]

                ref_hp_pos = seq_reference_start + ref_start

                # skip hps at the end of a segment since we don't know their true length
                if ref_hp_pos == seq_reference_start or ref_hp_pos + ref_hp_len == seq_reference_end:
                    continue

                if trees is not None:
                    # skip hps that at the end of an interval or spanning multiple intervals
                    # since we don't know their true length
                    interval_left_end = set(trees[seq.reference_name].find_overlap(ref_hp_pos - 1, ref_hp_pos))
                    interval_right_end = set(trees[seq.reference_name].find_overlap(ref_hp_pos + ref_hp_len, ref_hp_pos + ref_hp_len + 1))
                    if (len(interval_left_end) == 0 or len(interval_right_end) == 0 or
                        not bool(interval_left_end.intersection(interval_right_end))):
                        continue

                aligned_ref_len += ref_hp_len

                score += min((call_len - ref_hp_len) ** 2, 1000)
                out_str = '\t'.join([str(i) for i in (
                    seq.query_name,
                    seq.reference_name,
                    ref_hp_pos,
                    '-' if seq.is_reverse else '+',
                    query_start,
                    orginal_base,
                    base,
                    ref_hp_len,
                    call_len)])
                out_fh.write(out_str + '\n')
                # update counts of calls
                counts[base][ref_hp_len][call_len] += 1

    out_fh.close()

    return score, counts, aligned_ref_len


def main():
    """Entry point for homopolymer accuracy counting."""
    parser = argparse.ArgumentParser(
        prog='assess_homopolymers',
        description='Analyse homopolymer query and reference lengths.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command')
    subparsers.required = True

    cparser = subparsers.add_parser('count',
        help='Count homopolymers starting from a bam. ',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    cparser.set_defaults(func=count_homopolymers)
    cparser.add_argument('bam', help='Input bam file.')
    cparser.add_argument('-o', '--output_dir', default='homopolymers',
        help="Output directory (will be created).")
    cparser.add_argument('-t', '--threads', type=int, default=1,
        help='Number of threads for parallel execution.')
    cparser.add_argument('-l', '--homo_len', default=3, type=int,
        help='Minimum homopolymer length, default 3')
    cparser.add_argument('-b', '--bed',
        help='Bed file to limit search.')

    aparser = subparsers.add_parser('analyse',
        help='Analyse existing counts, optionally merging multiple counters.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    aparser.set_defaults(func=analyse_homopolymers)
    aparser.add_argument('pkl', nargs='+', help='Input .pkl file(s).')
    aparser.add_argument('-o', '--output_dir', default='homopolymers',
        help="Output directory (will be created).")

    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



"""
Tabulate some simple alignment stats from sam.
"""

import argparse
from collections import Counter
from concurrent.futures import ProcessPoolExecutor
import functools
import itertools
from math import ceil
import sys

import pysam

from pomoxis.util import intervaltrees_from_bed, masked_stats_from_aligned_read, stats_from_aligned_read

parser = argparse.ArgumentParser(
        prog='stats_from_bam',
        description="""Parse a bamfile (from a stream) and output summary stats for each read.""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('bam', type=str, help='Path to bam file.')
parser.add_argument('--bed', default=None, help='.bed file of reference regions to include.')
parser.add_argument('-m', '--min_length', type=int, default=None)
parser.add_argument('-a', '--all_alignments', action='store_true',
                    help='Include secondary and supplementary alignments.')
parser.add_argument('-o', '--output', type=argparse.FileType('w'), default=sys.stdout,
                    help='Output alignment stats to file instead of stdout.')
parser.add_argument('-s', '--summary', type=argparse.FileType('w'), default=sys.stderr,
                    help='Output summary to file instead of stderr.')
parser.add_argument('-t', '--threads', type=int, default=1,
                    help='Number of threads for parallel processing.')



def _process_reads(bam_fp, start_stop, all_alignments=False, min_length=None, bed_file=None):
    start, stop = start_stop
    counts = Counter()
    results = []
    lra_flag = False
    with pysam.AlignmentFile(bam_fp, 'rb') as bam:
        if bed_file is not None:
            trees = intervaltrees_from_bed(bed_file)
        for i, read in enumerate(bam.fetch(until_eof=True)):
            if i < start:
                continue
            elif i == stop:
                break
            if read.is_unmapped:
                counts['unmapped'] += 1
                continue
            if not all_alignments and (read.is_secondary or read.is_supplementary):
                continue

            lra_flag = False
            if bed_file is not None:
                if (read.reference_name not in trees or
                    not trees[read.reference_name].has_overlap(read.reference_start, read.reference_end)):
                    sys.stderr.write('read {} does not overlap with any regions in bedfile\n'.format(read.query_name))
                    counts['masked'] += 1
                    continue
                else:
                    result = masked_stats_from_aligned_read(read, trees[read.reference_name])
                    if result is None:  # no matches within bed regions
                        counts['all_matches_masked'] +=  1
                        continue
            else:
                result, lra_flag = stats_from_aligned_read(read)

            if min_length is not None and result['length'] < min_length:
                counts['short'] += 1
                continue

            results.append(result)

    return counts, results, lra_flag


def main(arguments=None):
    args = parser.parse_args(arguments)

    if args.threads > 1 and args.bed is not None:
        pool = ProcessPoolExecutor(args.threads)
        mapper = pool.map
    else:
        pool = None
        mapper = map
        args.threads = 1

    headers = ['name', 'ref', 'coverage', 'ref_coverage', 'qstart', 'qend',
               'rstart', 'rend', 'aligned_ref_len', 'direction', 'length',
               'read_length', 'mean_quality', 'match', 'ins', 'del', 'sub',
               'iden', 'acc', 'mapq', 'flag']

    masked_headers = ['masked']
    if args.bed is not None:
        headers.extend(masked_headers)

    args.output.write('\t'.join(headers))
    args.output.write('\n')

    # create a slice of reads to process in each thread to avoid looping through
    # bam n read times and reduce mp overhead
    ranges = [(0, float('inf'))]
    if args.threads > 1:
        with pysam.AlignmentFile(args.bam) as bam:
            n_reads = bam.count(until_eof=True)
        n_reads_per_proc = ceil(n_reads / args.threads)
        ranges = [(start, min(start + n_reads_per_proc, n_reads))
                   for start in range(0, n_reads, n_reads_per_proc)]

    func = functools.partial(_process_reads, args.bam,
                             all_alignments=args.all_alignments,
                             min_length=args.min_length, bed_file=args.bed)

    detected_lra = False
    counts = Counter()
    for batch_counts, results, lra_flag in mapper(func, ranges):
        counts.update(batch_counts)
        counts['total'] += len(results)
        for result in results:
            out_row = (str(result[x]) for x in headers)
            args.output.write('\t'.join(out_row))
            args.output.write('\n')
        detected_lra = detected_lra or lra_flag
    if pool is not None:
        pool.shutdown(wait=True)

    if detected_lra:
        args.summary.write('Some reads contained an NX tag, assuming these were aligned with LRA.\n')
        args.summary.write('If this is not the case, results may be wrong.\n')

    if counts['total'] == 0:
        args.summary.write('No alignments processed. Check your bam and filtering options.\n')

    args.summary.write('Mapped/Unmapped/Short/Masked/Skipped(all matches masked): {total}/{unmapped}/{short}/{masked}/{all_matches_masked}\n'.format_map(counts))


if __name__ == '__main__':
    main()



import argparse
import functools
import logging
import os

import numpy as np
import pandas as pd
import pysam

from pomoxis.util import filter_args, filter_read, parse_regions, Region


def coverage_of_region(region, bam_fp, stride, read_filter_func):
    """Get coverage data for a region"""

    bins = np.arange(region.start, region.end, stride)
    msg = 'Processing region {}:{}-{}'
    logging.info(msg.format(region.ref_name, bins[0], bins[-1]))
    coverage_by_is_rev = {True: np.zeros(len(bins)), False: np.zeros(len(bins))}
    for r_obj in pysam.AlignmentFile(bam_fp).fetch(contig=region.ref_name, start=region.start, end=region.end):
        # Ignore secondary/supplementary maps when computing the median depth, if requested
        if read_filter_func(r_obj):
            continue
        start_i = max((r_obj.reference_start - bins[0]) // stride, 0)
        end_i = min((r_obj.reference_end - bins[0]) // stride, len(bins))
        coverage_by_is_rev[r_obj.is_reverse][start_i: end_i] += 1
    total_depth = coverage_by_is_rev[True] + coverage_by_is_rev[False]

    df = pd.DataFrame({'pos': bins,
                       'depth': total_depth,
                       'depth_fwd': coverage_by_is_rev[False],
                       'depth_rev': coverage_by_is_rev[True],
                       }
    )
    return df


def coverage_summary_of_region(*args, **kwargs):
    df = coverage_of_region(*args, **kwargs)
    return df.describe()


def main():
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)

    parser = argparse.ArgumentParser(
        prog='coverage_from_bam',
        description='Calculate read coverage depth from a bam.',
        epilog='By default a file is written per reference sequence, this can be changed with the `--one_file` '
        'option. If overlapping regions are specified, `--one_file` should not be used.',
        parents=[filter_args()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bam', help='.bam/sam file.')
    parser.add_argument('-r', '--regions', nargs='+', help='Only process given regions.')
    grp = parser.add_mutually_exclusive_group()
    grp.add_argument('-p', '--prefix', help='Prefix for output, defaults to basename of bam.')
    grp.add_argument('-o', '--one_file', help='Single output file with "region" column.')
    parser.add_argument('-s', '--stride', type=int, default=1000, help='Stride in genomic coordinate.')
    parser.add_argument('--summary_only', action='store_true',
                        help='Output only the depth_summary.txt file')

    args = parser.parse_args()
    read_filter_func = functools.partial(filter_read, args=args)

    bam = pysam.AlignmentFile(args.bam)
    ref_lengths = dict(zip(bam.references, bam.lengths))

    if args.regions is not None:
        regions = parse_regions(args.regions, ref_lengths=ref_lengths)
    else:
        regions = [Region(ref_name=r, start=0, end=ref_lengths[r]) for r in bam.references]

    summary = {}
    first = True
    if args.one_file is not None:
        args.one_file = open(args.one_file, 'w')
    for region in regions:
        prefix = args.prefix
        if prefix is None:
            prefix = os.path.splitext(os.path.basename(args.bam))[0]

        region_str = '{}_{}_{}'.format(region.ref_name, region.start, region.end)
        df = coverage_of_region(region, args.bam, args.stride, read_filter_func)
        summary[region_str] = df['depth'].describe()

        if not args.summary_only:
            if args.one_file is not None:
                df['rname'] = region.ref_name
                df.to_csv(args.one_file, sep='\t', index=False, header=first)
            else:
                depth_fp = '{}_{}.depth.txt'.format(prefix, region_str)
                df.to_csv(depth_fp, sep='\t', index=False)
        first = False
    if args.one_file is not None:
        args.one_file.close()


    summary_fp = '{}_depth_summary.txt'.format(prefix)
    summary_df = pd.DataFrame(summary).T.reset_index().rename(columns={'index': 'region'})
    summary_df.to_csv(summary_fp, index=False, sep='\t')


if __name__ == '__main__':
    main()



__version__ = '0.3.16'

import argparse
import os
import sys
import platform
import subprocess

def get_prog_path(prog):
    """Get the absolute path of bundled executables.

    :param prog: programme (file) name.

    :returns: absolute path to executable.
    """
    prog_path = os.path.join(
        os.path.dirname(__file__), '..', 'exes', prog
    )
    if not (os.path.isfile(prog_path) and os.access(prog_path, os.X_OK)):
        raise RuntimeError('Cannot find executable "{}".'.format(prog))
    return prog_path


def show_prog_path():
    """Print the path of bundled executables."""
    parser = argparse.ArgumentParser(
        prog='pomoxis_path',
        description='Print the path of bundled executables.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('program', help='name of program.')
    args = parser.parse_args()
    print(get_prog_path(args.program))


def run_prog(prog, args, stdout=None):
    """Run one of the bundled executables.

    :param prog: programme name.
    :param args: commandline arguments for programme.
    :param stdout: filehandle/path for stdout of programme.

    :returns: result of `subprocess.call`.
    """
    prog_path = get_prog_path(prog)
    if stdout is None:
        return subprocess.check_output([prog_path] + args)
    else:
        with open(stdout, 'w') as fh:
            return subprocess.call([prog_path] + args, stdout=fh, stderr=open(os.devnull, 'w'))




import argparse
from collections import namedtuple, defaultdict
import itertools
import logging
import os
import shutil
import sys

from ncls import NCLS
import numpy as np
import pandas as pd
import pysam

Region = namedtuple('Region', 'ref_name start end')
AlignPos = namedtuple('AlignPos', ('qpos', 'qbase', 'rpos', 'rbase'))


class FastxWrite:
    def __init__(self, fname, mode='w', width=80, force_q=False, mock_q=10):
        self.fname = fname
        self.mode = mode
        self.width = width
        self.force_q = force_q
        self.mock_q = chr(33 + mock_q)
        self.aq = None

    def __enter__(self):
        if self.fname == '-':
            self.fh = sys.stdout
        else:
            self.fh = open(self.fname, self.mode)
        return self

    def __exit__(self, type, value, traceback):
        if self.fname != '-':
            self.fh.close()

    def write(self, name, seq, qual=None, comment=''):
        if comment != '':
            comment = ' {}'.format(comment)
        if (self.aq == 'a' and qual is not None) or \
                (self.aq == 'q' and qual is None):
            raise ValueError("Type of records changed whilst writing.")
        if qual or self.force_q:
            if qual is None:
                qual = self.mock_q * len(seq)
            self.aq = 'q'
            self.fh.write("@{}{}\n{}\n+\n{}\n".format(name, comment, seq, qual))
        else:
            self.aq = 'a'
            self.fh.write(">{}{}\n".format(name, comment))
            for _, chunk in enumerate(chunks(seq, self.width)):
                self.fh.write('{}\n'.format(''.join(chunk)))


def chunks(iterable, n):
    """Generate fixed length chunks of an interable.

    :param iterable: input sequence.
    :param n: chunk size.
    """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


def cat(files, output, chunks=1024*1024*10):
    """Concatenate a set of files.

    :param files: input filenames.
    :param output: output filenames.
    :param chunks: buffersize for filecopy.
    """
    with open(output, 'wb') as wfd:
        for f in files:
            with open(f, 'rb') as fd:
                shutil.copyfileobj(fd, wfd, chunks)


def split_fastx(fname, output, chunksize=10000):
    """Split records in a fasta/q into fixed lengths.

    :param fname: input filename.
    :param output: output filename.
    :param chunksize: (maximum) length of output records.
    """
    with FastxWrite(output, 'w') as fout:
        with pysam.FastxFile(fname, persist=False) as fin:
            for rec in fin:
                name = rec.name
                seq = rec.sequence
                qual = rec.quality
                if rec.comment is None:
                    comment = 'chunk_length={}'.format(chunksize)
                else:
                    comment = '{} chunk_length={}'.format(rec.comment, chunksize)
                if qual is None:
                    for i, s in enumerate(chunks(seq, chunksize)):
                        chunk_name = '{}_chunk{}'.format(name, i)
                        fout.write(chunk_name, s, comment=comment)
                else:
                    for i, (s, q) in enumerate(zip(chunks(seq, chunksize), chunks(qual, chunksize))):
                        chunk_name = '{}_chunk{}'.format(name, i)
                        fout.write(chunk_name, ''.join(s), ''.join(q), comment)


def split_fastx_cmdline():
    """Split records in a fasta/q file into chunks of a maximum size."""
    parser = argparse.ArgumentParser(
        prog='split_fastx',
        description='Split records in a fasta/q file into chunks of a maximum size.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input', help='Input fastax/q file.')
    parser.add_argument('output', help='Output fastax/q file.')
    parser.add_argument('chunksize', type=int, help='Maximum size of output sequences.')
    args = parser.parse_args()
    split_fastx(args.input, args.output, args.chunksize)


def fast_convert():
    """Convert between fasta<->fastq."""
    parser = argparse.ArgumentParser(
        prog='fast_convert',
        description='Convert between fasta<->fastq.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('convert', choices=['qq', 'aa', 'aq', 'qa'],
        help='Conversion code: from->to.')
    parser.add_argument('--discard_q', action='store_true',
        help='Discard quality information from fastq, use with --mock_q.')
    parser.add_argument('--mock_q', default=10, type=int,
        help='Mock quality value, valid for convert=aq|qq.')
    args = parser.parse_args()

    in_fmt = 'fastq'
    out_fmt= 'fasta'
    qflag = False # controls quality manipulation
    if args.convert == 'qq':
        out_fmt = 'fastq'
        if args.discard_q is not None:
            qflag = True
    elif args.convert == 'aa':
        in_fmt = 'fasta'
    elif args.convert == 'aq':
        in_fmt = 'fasta'
        out_fmt = 'fastq'
        qflag = True
    elif args.convert == 'qa':
        pass # default
    else:
        raise ValueError("convert must be 'qq', 'aq', 'qa', or 'aa'\n")

    with FastxWrite('-', force_q=out_fmt=='fastq', mock_q=args.mock_q) as fh_out:
        with pysam.FastxFile('-') as fh_in:
            for rec in fh_in:
                if out_fmt == 'fasta':
                    qual = None
                elif qflag:
                    qual = chr(33 + args.mock_q) * len(rec.sequence)
                else:
                    qual = rec.quality

                fh_out.write(rec.name, rec.sequence, qual, rec.comment)


def extract_long_reads():
    """Filter fastq to longest reads."""

    parser = argparse.ArgumentParser(description='Extract longest reads from a fastq.')
    parser.add_argument('input',
        help='Input .fasta/q file.')
    parser.add_argument('output',
        help='Output .fasta file.')
    filt = parser.add_mutually_exclusive_group(required=True)
    filt.add_argument('--longest', default=None, type=float,
        help='Percentage of longest reads to partition.')
    filt.add_argument('--bases', default=None, type=int,
        help='Maximum number of bases (subject to at least one read.)')
    parser.add_argument('--others', default=None,
        help='Write all other reads to file.')
    args = parser.parse_args()

    sys.stderr.write('Loading reads...\n')
    record_dict = dict()
    for rec in pysam.FastxFile(args.input):
        record_dict[rec.name] = (rec.sequence, rec.quality, rec.comment)

    ids = list(record_dict.keys())
    lengths = np.fromiter(
        (len(record_dict[i][0]) for i in ids),
        dtype=int, count=len(ids)
    )
    sys.stderr.write('Sorting reads...\n')
    if args.bases is None:
        # partial sort will do fine here
        max_reads = int(len(ids) * (args.longest / 100))
        longest = np.argpartition(lengths, -max_reads)[-max_reads:]
    else:
        # need a full sort
        order = np.argsort(lengths)[::-1]
        cumsum = 0
        last = 1
        for i, j in enumerate(order, 1):
            cumsum += lengths[j]
            if cumsum > args.bases:
                break
            last = i
        longest = order[:last]

    with FastxWrite(args.output, 'w') as out:
        for i in longest:
            name = ids[i]
            rec = record_dict[ids[i]]
            out.write(name, *rec)

    if args.others is not None:
        if args.bases is None:
            others = np.argpartition(lengths, -max_reads)[:-max_reads]
        else:
            others = order[last:]

        with FastxWrite(args.others, 'w') as out:
            for i in others:
                name = ids[i]
                rec = record_dict[ids[i]]
                out.write(name, *rec)


def parse_regions(regions, ref_lengths=None):
    """Parse region strings into `Region` objects.

    :param regions: iterable of str
    :param ref_lengths: {str ref_names: int ref_lengths}, if provided Region.end
        will default to the reference length instead of None.

    >>> parse_regions(['Ecoli'])[0]
    Region(ref_name='Ecoli', start=0, end=None)
    >>> parse_regions(['Ecoli:1000-2000'])[0]
    Region(ref_name='Ecoli', start=1000, end=2000)
    >>> parse_regions(['Ecoli:-1000'])[0]
    Region(ref_name='Ecoli', start=0, end=1000)
    >>> parse_regions(['Ecoli:500-'])[0]
    Region(ref_name='Ecoli', start=500, end=None)
    >>> parse_regions(['Ecoli'], ref_lengths={'Ecoli':4800000})[0]
    Region(ref_name='Ecoli', start=0, end=4800000)
    >>> parse_regions(['NC_000921.1:10000-20000'])[0]
    Region(ref_name='NC_000921.1', start=10000, end=20000)
    """
    def _get_end(end, ref_name):
        return end if end is not None else (
            ref_lengths[ref_name] if ref_lengths is not None else None
        )

    decoded = []
    for region in regions:
        if ':' not in region:
            ref_name, start, end = region, 0, None
        else:
            start, end = 0, None
            ref_name, bounds = region.split(':')
            if bounds[0] == '-':
                start = 0
                end = int(bounds[1:])
            elif bounds[-1] == '-':
                start = int(bounds[:-1])
                end = None
            else:
                start, end = [int(b) for b in bounds.split('-')]
        decoded.append(Region(ref_name, start, _get_end(end, ref_name)))
    return tuple(decoded)


class SeqLen(argparse.Action):
    """Parse a sequence length from str such as 4.8mb or from fastx."""
    def __call__(self, parser, namespace, values, option_string=None):
        seq_len = None
        try:
            seq_len = int(values)
        except:
            suffixes = {
                'kb': 10 ** 3,
                'mb': 10 ** 6,
                'gb': 10 ** 9,
            }
            for suffix, multiplier in suffixes.items():
                if suffix in values.lower():
                    seq_len = int(multiplier * float(values.lower().replace(suffix, '')))
                    break
            if seq_len is None:
                try:
                    seq_len = sum(get_seq_lens(values))
                except:
                    raise ValueError('Could not get sequence length from {}.'.format(values))

        setattr(namespace, self.dest, seq_len)


def get_seq_lens(fastx):
    """Get sequence lengths from fastx file"""
    return [len(r.sequence) for r in pysam.FastxFile(fastx)]


def coverage_from_fastx():
    parser = argparse.ArgumentParser(
        prog='coverage_from_fastx',
        description='Estimate coverage from summed basecall and reference lengths',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('basecalls', help='input fastx file.')
    parser.add_argument('ref_len', action=SeqLen,
                        help='reference length (e.g. 4.8kb/mb/gb) or reference fastx from which to calculate length.')
    parser.add_argument('--coverage', type=int, help='Calculate fraction of reads required for this coverage.')
    parser.add_argument('--longest', action='store_true', default=False,
                        help='Use the longest reads when calculating fraction reads required for a given coverage.')

    args = parser.parse_args()

    basecall_lens = get_seq_lens(args.basecalls)
    total_basecalls_len = sum(basecall_lens)

    print('Total length of basecalls: {}'.format(total_basecalls_len))
    print('Total ref length: {}'.format(args.ref_len))
    print('Estimated coverage: {:.2f}'.format(total_basecalls_len / args.ref_len))

    if args.coverage is not None:
        if args.longest:
            basecall_lens.sort(reverse=True)
            cum_len = np.cumsum(basecall_lens)
            required_len = args.coverage * args.ref_len
            n_required = np.searchsorted(cum_len, required_len, side='right')
            msg = '% of longest reads required to achieve {}X coverage: {}'
            print(msg.format(args.coverage, 100 * n_required / len(cum_len)))
        else:
            msg = '% of randomly-selected reads required to achieve {}X coverage: {}'
            print(msg.format(args.coverage, 100 * args.coverage  * args.ref_len / total_basecalls_len))


def get_pairs(aln):
    """Return generator of pairs.

    :param aln: `pysam.AlignedSegment` object.
    :returns: generator of `AlignPos` objects.
    """
    seq = aln.query_sequence
    pairs = (AlignPos(qpos=qp,
                      qbase=seq[qp] if qp is not None else '-',
                      rpos=rp,
                      rbase=rb if rp is not None else '-'
                      )
             for qp, rp, rb in aln.get_aligned_pairs(with_seq=True)
             )
    return pairs


def get_trimmed_pairs(aln):
    """Trim aligned pairs to the alignment.

    :param aln: `pysam.AlignedSegment` object
    :yields pairs:
    """

    pairs = get_pairs(aln)
    for pair in itertools.dropwhile(lambda x: x.rpos is None or x.qpos is None, pairs):
        if (pair.rpos == aln.reference_end or pair.qpos == aln.query_alignment_end):
            break
        yield pair


def yield_from_bed(bedfile):
    with open(bedfile) as fh:
        for line in fh:
            if line.lstrip()[0] == '#':
                continue
            split_line = line.split()
            if split_line[0] in {'browser', 'track'} or len(split_line) < 3:
                continue
            chrom = split_line[0]
            start = int(split_line[1])
            stop = int(split_line[2])
            yield chrom, start, stop


def intervaltrees_from_bed(path_to_bed):
    """Created dict of intervaltrees from a .bed file, indexed by chrom.

    :param path_to_bed: str, path to .bed file.
    :returns: { str chrom: `ncls.NCLS` obj }.
    """
    tmptrees = defaultdict(list)
    trees = {}
    for chrom, start, stop in yield_from_bed(path_to_bed):
        tmptrees[chrom].append([start, stop])
    for chrom, d in tmptrees.items():
        a = np.array(d)
        trees[chrom] = NCLS(a[:, 0], a[:, 1], np.arange(a.shape[0]))
    return trees


def tag_bam():
    """Command line tool to add tags to a bam."""
    parser = argparse.ArgumentParser(
        prog='tag_bam',
        description='Add a tag to all alignments in a bam.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input', help='Input bam file.')
    parser.add_argument('output', help='Output output file.')
    parser.add_argument('tag_name', help='Tag name.')
    parser.add_argument('tag_value', type=int, help='Tag value.')
    args = parser.parse_args()
    with pysam.AlignmentFile(args.input) as bam_in:
        with pysam.AlignmentFile(args.output, 'wb', header=bam_in.header) as bam_out:
            for r in bam_in:
                r.set_tag(args.tag_name, args.tag_value)
                bam_out.write(r)


def reverse_bed():
    """Convert bed-file coordinates to coordinates on the reverse strand."""
    parser = argparse.ArgumentParser(
        prog='reverse_bed',
        description='Convert bed-file coordinates to coordinates on the reverse strand.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('bed_in', help='Input bed file.')
    parser.add_argument('ref_fasta', help='Input reference fasta file.')
    parser.add_argument('bed_out', help='Output bed file.')
    args = parser.parse_args()

    fasta = pysam.FastaFile(args.ref_fasta)
    lengths = dict(zip(fasta.references, fasta.lengths))
    d = pd.read_csv(args.bed_in, sep='\t', names=['chrom', 'start', 'stop'])

    d['chrom_length'] = d['chrom'].map(lambda x: lengths[x])
    d['rc_stop'] = d['chrom_length'] - d['start']
    d['rc_start'] = d['chrom_length'] - d['stop']
    d['chrom_rc'] = d['chrom'] + '_rc'
    d[['chrom_rc', 'rc_start', 'rc_stop']].to_csv(args.bed_out, index=False, header=False, sep='\t')


class PrimSupAction(argparse.Action):
    """Parse primary / supplementary option."""

    def __call__(self, parser, namespace, values, option_string=None):
        if self.dest == 'primary_only':
            setattr(namespace, self.dest, True)
            setattr(namespace, 'keep_supplementary', False)
        elif self.dest == 'keep_supplementary':
            setattr(namespace, self.dest, True)
            setattr(namespace, 'primary_only', False)


def stats_from_aligned_read(read):
    """Create summary information for an aligned read

    :param read: :class:`pysam.AlignedSegment` object
    """
    try:
        NM = read.get_tag('NM')
    except:
        raise IOError("Read is missing required 'NM' tag. Try running 'samtools fillmd -S - ref.fa'.")

    name = read.query_name
    start_offset = 0
    if read.is_secondary or read.is_supplementary:
        first_cig = read.cigartuples[0]
        if first_cig[0] == 5:  # should always be true for minimap2
            start_offset = first_cig[1]
    counts, _ = read.get_cigar_stats()
    match = counts[0] + counts[7] + counts[8]  # total of M, =, and X
    ins = counts[1]
    delt = counts[2]

    lra_flag = False
    if read.has_tag('NX'):
        # likely from lra
        # NM is number of matches, see https://github.com/ChaissonLab/LRA/issues/32
        sub = counts[8]
        lra_flag = True
    else:
        # likely from minimap2
        # NM is edit distance: NM = INS + DEL + SUB
        sub = NM - ins - delt

    length = match + ins + delt
    iden = 100 * float(match - sub) / match
    acc = 100 * float(match - sub) / length

    read_length = read.infer_read_length()
    coverage = 100 * float(read.query_alignment_length) / read_length
    direction = '-' if read.is_reverse else '+'

    if read.query_qualities is None:
        mean_quality = None
    else:
        mean_quality = round(mean_q_from_qualities(read.query_qualities), ndigits=2)

    results = {
        "name": name,
        "coverage": coverage,
        "direction": direction,
        "length": length,
        "read_length": read_length,
        "mean_quality": mean_quality,
        "match": match,
        "ins": ins,
        "del": delt,
        "sub": sub,
        "iden": iden,
        "acc": acc,
        "qstart": read.query_alignment_start + start_offset,
        "qend": read.query_alignment_end + start_offset,
        "rstart": read.reference_start,
        "rend": read.reference_end,
        "ref": read.reference_name,
        "aligned_ref_len": read.reference_length,
        "ref_coverage": 100*float(read.reference_length) / read.header.lengths[read.reference_id],
        "mapq": read.mapping_quality,
        "flag": read.flag,
    }

    return results, lra_flag


def masked_stats_from_aligned_read(read, tree):
    """Create summary information for an aligned read over regions in bed file.

    :param read: :class:`pysam.AlignedSegment` object
    """
    try:
        MD = read.get_tag('MD')
    except:
        raise IOError("Read is missing required 'MD' tag. Try running 'samtools callmd - ref.fa'.")

    correct, delt, ins, sub, aligned_ref_len, masked = 0, 0, 0, 0, 0, 0
    pairs = read.get_aligned_pairs(with_seq=True)
    qseq = read.query_sequence
    pos_is_none = lambda x: (x[1] is None or x[0] is None)
    pos = None
    insertions = []
    # TODO: refactor to use get_trimmed_pairs (as in catalogue_errors)?
    for qp, rp, rb in itertools.dropwhile(pos_is_none, pairs):
        if rp == read.reference_end or (qp == read.query_alignment_end):
            break
        pos = rp if rp is not None else pos
        if not tree.has_overlap(pos, pos + 1) or (rp is None and not tree.has_overlap(pos + 1, pos + 2)):
            # if rp is None, we are in an insertion, check if pos + 1 overlaps
            # (ref position of ins is arbitrary)
            # print('Skipping ref {}:{}'.format(read.reference_name, pos))
            masked += 1
            continue
        else:
            if rp is not None:  # we don't have an insertion
                aligned_ref_len += 1
                if qp is None:  # deletion
                    delt += 1
                elif qseq[qp] == rb:  # correct
                    correct += 1
                elif qseq[qp] != rb:  # sub
                    sub += 1
            else:  # we have an insertion
                ins += 1

    name = read.query_name
    match = correct + sub
    length = match + ins + delt

    if match == 0:
        # no matches within bed regions - all bed ref positions were deleted.
        # skip this alignment.
        return None

    iden = 100 * float(match - sub) / match
    acc = 100 - 100 * float(sub + ins + delt) / length

    read_length = read.infer_read_length()
    masked_query_alignment_length = correct + sub + ins
    coverage = 100*float(masked_query_alignment_length) / read_length
    direction = '-' if read.is_reverse else '+'

    if read.query_qualities is None:
        mean_quality = None
    else:
        mean_quality = round(mean_q_from_qualities(read.query_qualities), ndigits=2)


    results = {
        "name": name,
        "coverage": coverage,
        "direction": direction,
        "length": length,
        "read_length": read_length,
        "mean_quality": mean_quality,
        "match": match,
        "ins": ins,
        "del": delt,
        "sub": sub,
        "iden": iden,
        "acc": acc,
        "qstart": read.query_alignment_start,
        "qend": read.query_alignment_end,
        "rstart": read.reference_start,
        "rend": read.reference_end,
        "ref": read.reference_name,
        "aligned_ref_len": aligned_ref_len,
        "ref_coverage": 100 * float(aligned_ref_len) / read.header.lengths[read.reference_id],
        "mapq": read.mapping_quality,
        "flag": read.flag,
        "masked": masked,
    }
    return results


def filter_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    group = parser.add_argument_group('Read filtering options')
    group.add_argument('-O', '--orientation', choices=['fwd', 'rev'],
        help='Sample only forward or reverse reads.')
    group.add_argument('-q', '--quality', type=float,
        help='Filter reads by mean qscore.')
    group.add_argument('-a', '--accuracy', type=float,
        help='Filter reads by accuracy.')
    group.add_argument('-c', '--coverage', type=float,
        help='Filter reads by coverage (what fraction of the read aligns).')
    group.add_argument('-l', '--length', type=int, default=None,
        help='Filter reads by read length.')
    group.add_argument('--keep_unmapped', action='store_true',
        help='Include unmapped reads.')

    pgroup = group.add_mutually_exclusive_group()
    pgroup.add_argument('--primary_only', action=PrimSupAction, default=True, nargs=0,
                        help='Use only primary reads.')
    pgroup.add_argument('--keep_supplementary', action=PrimSupAction, default=False, nargs=0,
        help='Include supplementary alignments.')
    return parser


QSCORES_TO_PROBS = 10 ** (-0.1 * np.array(np.arange(100)))

def mean_q_from_qualities(qualities):
    return -10 * np.log10(np.mean(QSCORES_TO_PROBS[qualities]))


def filter_read(r, args, logger=None):
    """Decide whether a read should be filtered out, returning a bool"""

    if logger is None:
        logger = logging.getLogger('Filter')
    # filter secondary and unmapped reads
    if r.is_secondary:
        return True
    if r.is_unmapped and not args.keep_unmapped:
        return True
    if r.is_supplementary and not args.keep_supplementary:
        return True

    # Filters that apply on unaligned reads
    # filter quality
    if args.quality is not None:
        mean_q = mean_q_from_qualities(r.query_qualities)
        if mean_q < args.quality:
            logger.debug(f"Filtering {r.query_name} with quality {mean_q:.2f}")
            return True

    # filter length
    if args.length is not None:
        read_length = r.query_length
        if read_length < args.length:
            logger.info("Filtering {} by length ({:.2f}).".format(r.query_name, read_length))
            return True

    # Filters that only apply on aligned reads
    if not r.is_unmapped:
        # filter orientation
        if (r.is_reverse and args.orientation == 'fwd') or \
                (not r.is_reverse and args.orientation == 'rev'):
            return True

        if args.accuracy is not None or args.coverage is not None:
            stats, _ = stats_from_aligned_read(r)
            if args.accuracy is not None and stats['acc'] < args.accuracy:
                logger.info("Filtering {} by accuracy ({:.2f}).".format(r.query_name, stats['acc']))
                return True
            if args.coverage is not None and stats['coverage'] < args.coverage:
                logger.info("Filtering {} by coverage ({:.2f}).".format(r.query_name, stats['coverage']))
                return True
    # don't filter
    return False


def write_bam(bam, prefix, region, sequences, keep_supplementary=False):
    """Write subset bam.

    :param bam: str, path to input bam.
    :param prefix: str, prefix for output bam.
    :param bam: `Region` obj of region to subset to.
    :param sequences: set of query names to retain.
    :param keep_supplementary: bool, whether to retain supplementary alignments.
    """
    # filtered bam
    sequences = set(sequences)
    taken = set()
    output = '{}_{}.{}'.format(prefix, region.ref_name, os.path.basename(bam))
    src_bam = pysam.AlignmentFile(bam, "rb")
    out_bam = pysam.AlignmentFile(output, "wb", template=src_bam)
    for read in src_bam.fetch(region.ref_name, region.start, region.end):
        if read.is_secondary:
            continue
        if read.is_supplementary and not keep_supplementary:
            continue
        if read.query_name in sequences and (read.query_name, read.is_supplementary) not in taken:
            out_bam.write(read)
            taken.add((read.query_name, read.is_supplementary))
    src_bam.close()
    out_bam.close()
    pysam.index(output)



from glob import glob
import os
import re
from setuptools import setup, find_packages
from setuptools import Distribution
from setuptools.command.install import install
import shutil
import sys

__pkg_name__ = 'pomoxis'
__author__ = 'cwright'
__description__ = 'Assembly, consensensus, and analysis tools by ONT research.'

# Use readme as long description and say its github-flavour markdown
from os import path
this_directory = path.abspath(path.dirname(__file__))
kwargs = {'encoding':'utf-8'} if sys.version_info.major == 3 else {}
with open(path.join(this_directory, 'README.md'), **kwargs) as f:
    __long_description__ = f.read()
__long_description_content_type__ = 'text/markdown'

__path__ = os.path.dirname(__file__)
__pkg_path__ = os.path.join(os.path.join(__path__, __pkg_name__))

# Get the version number from __init__.py, and exe_path
verstrline = open(os.path.join(__pkg_name__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_name__))

dir_path = os.path.dirname(__file__)
install_requires = []
with open(os.path.join(dir_path, 'requirements.txt')) as fh:
    reqs = (
        r.split('#')[0].strip()
        for r in fh.read().splitlines() if not r.startswith('#')
    )
    for req in reqs:
        if req == '':
            continue
        if req.startswith('git+https'):
            req = req.split('/')[-1].split('@')[0]
        install_requires.append(req)

data_files = []
extensions = []
extra_requires={}

# Read these from Makefile in the event someone modified list there
exes = []
if os.environ.get("POMO_BINARIES") is not None:
    with open(os.path.join(dir_path, 'Makefile')) as fh:
        for line in fh.readlines():
            tokens = line.split('=')
            if tokens[0] == 'BINARIES':
                exes = tokens[1].split()
                break
    #place binaries as package data, below we'll copy them to standard path in dist
    data_files.append(
        ('exes', [
            'bincache/{}'.format(x, x) for x in exes
        ])
    )

setup(
    name=__pkg_name__,
    version=__version__,
    author=__author__,
    author_email='{}@nanoporetech.com'.format(__author__),
    description=__description__,
    long_description=__long_description__,
    long_description_content_type=__long_description_content_type__,
    dependency_links=[],
    ext_modules=extensions,
    install_requires=install_requires,
    tests_require=[].extend(install_requires),
    extras_require=extra_requires,
    python_requires='>=3.6',
    packages=find_packages(exclude=['*.test', '*.test.*', 'test.*', 'test']),
    package_data={},
    zip_safe=False,
    test_suite='discover_tests',
    data_files=data_files,
    entry_points={
        'console_scripts': [
            'catalogue_errors = {}.catalogue_errors:main'.format(__pkg_name__),
            'common_errors_from_bam = {}.common_errors_from_bam:main'.format(__pkg_name__),
            'assess_homopolymers = {}.assess_homopolymers:main'.format(__pkg_name__),
            'coverage_from_bam = {}.coverage_from_bam:main'.format(__pkg_name__),
            'coverage_from_fastx = {}.util:coverage_from_fastx'.format(__pkg_name__),
            'fast_convert = {}.util:fast_convert'.format(__pkg_name__),
            'long_fastx = {}.util:extract_long_reads'.format(__pkg_name__),
            'pomoxis_path = {}:show_prog_path'.format(__pkg_name__),
            'qscores_from_summary = {}.qscores_from_summary:main'.format(__pkg_name__),
            'reverse_bed = {}.util:reverse_bed'.format(__pkg_name__),
            'split_fastx = {}.util:split_fastx_cmdline'.format(__pkg_name__),
            'stats_from_bam = {}.stats_from_bam:main'.format(__pkg_name__),
            'subsample_bam = {}.subsample_bam:main'.format(__pkg_name__),
            'filter_bam = {}.filter_bam:main'.format(__pkg_name__),
            'summary_from_stats = {}.summary_from_stats:main'.format(__pkg_name__),
            'tag_bam = {}.util:tag_bam'.format(__pkg_name__),
            'trim_alignments = {}.trim_alignments:main'.format(__pkg_name__),
            'ref_seqs_from_bam = {}.ref_seqs_from_bam:main'.format(__pkg_name__),
            'find_indels = {}.find_indels:main'.format(__pkg_name__),
        ] # leave this here
    },
    scripts=[
        'scripts/assess_assembly',
        'scripts/intersect_assembly_errors',
        'scripts/mini_align',
        'scripts/mini_assemble']
)


# Nasty hack to get binaries into bin path
class GetPaths(install):
    def run(self):
        self.distribution.install_scripts = self.install_scripts
        self.distribution.install_libbase = self.install_libbase

def get_setuptools_script_dir():
    # Run the above class just to get paths
    dist = Distribution({'cmdclass': {'install': GetPaths}})
    dist.dry_run = True
    dist.parse_config_files()
    command = dist.get_command_obj('install')
    command.ensure_finalized()
    command.run()

    src_dir = glob(os.path.join(dist.install_libbase, 'pomoxis-*', 'exes'))[0]
    for exe in (os.path.join(src_dir, x) for x in os.listdir(src_dir)):
        print("Copying", os.path.basename(exe), '->', dist.install_scripts)
        shutil.copy(exe, dist.install_scripts)
    return dist.install_libbase, dist.install_scripts

if os.environ.get("POMO_BINARIES") is not None:
    print("\nCopying utility binaries to your path.")
    get_setuptools_script_dir()




import numpy as np
from pysam import FastaFile

"""Bioinformatics helpers."""

comp = {
    'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 'X': 'X', 'N': 'N',
    'a': 't', 't': 'a', 'c': 'g', 'g': 'c', 'x': 'x', 'n': 'n',
    #'-': '-'
}
comp_trans = str.maketrans(''.join(comp.keys()), ''.join(comp.values()))


def reverse_complement(seq):
    """Reverse complement sequence.

    :param: input sequence string.

    :returns: reverse-complemented string.
    """
    return seq.translate(comp_trans)[::-1]


def shotgun_library(fasta_file, mu, sigma, direction=(1,-1)):
    """Generate random fragment sequences of a given input sequence

    :param seq: input sequence.
    :param mu: mean fragment length.
    :param sigma: stdv of fragment length.
    :param direction: tuple represention direction of output sequences with
        respect to the input sequence.

    :yields: sequence fragments.

    .. note:: Could be made more efficient using buffers for random samples
        and handling cases separately.
    """
    fasta = FastaFile(fasta_file)
    seq_lens = [fasta.get_reference_length(x) for x in fasta.references]
    total_len = sum(seq_lens)
    seq_probs = [x / total_len for x in seq_lens]
    # FastaFile.fetch is proper slow, just read everything
    refs = fasta.references
    fasta = {k:fasta.fetch(k) for k in refs}

    def random_buffer(probs, size=10000):
        while True:
            buf = []
            for x, n in zip(range(len(probs)), np.random.multinomial(size, probs)):
                buf.extend([x]*n)
            np.random.shuffle(buf)
            for x in buf:
                yield x
    seq_chooser = random_buffer(seq_probs)
   
    # parameters for lognormal
    mean = np.log(mu / np.sqrt(1 + sigma**2 / mu**2))
    stdv = np.sqrt(np.log(1 + sigma**2 / mu**2))
    
    while True:
        # choose a seq based on length
        seq_i = next(seq_chooser)
        seq = fasta[refs[seq_i]]
        seq_len = seq_lens[seq_i]

        start = np.random.randint(0, seq_len)
        frag_length = int(np.random.lognormal(mean, stdv))
        move = np.random.choice(direction)
        end = max(0, start + move*frag_length)
        end = min(end, seq_len)
        start, end = sorted([start, end])

        if end - start < 2:
            # Expand a bit to ensure we grab at least one base.
            start = max(0, start - 1)
            end += 1

        frag_seq = seq[start:end]
        if move == -1:
            frag_seq = reverse_complement(frag_seq)
        yield frag_seq, refs[seq_i], start, end, '+' if move == 1 else '-'




import argparse
import os
import pysam
from Bio import SeqIO
from Bio.Seq import Seq


def main():
    parser = argparse.ArgumentParser(
        prog='trim_alignments',
        description='Trim alignments in multiple bams to common overlap window.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bams', nargs='+', help='input bam files')
    parser.add_argument('-r', '--ref_name',
                        help='Reference to process, only needed if bams contain multiple references.')
    parser.add_argument('-o', '--output_prefix', default='trimmed',
                        help='Prefix for outputs.')
    parser.add_argument('-f', '--reference_fasta', default=None,
                        help='Reference fasta to trim to alignment window.')

    args = parser.parse_args()

    refs = set()
    for bam in args.bams:
        with pysam.AlignmentFile(bam) as b:
            refs.update(b.references)

    if args.ref_name is None:
        if len(refs) > 1:
            raise RuntimeError('Bams contain multiple references, '
                               +'use the --ref_name argument to specify one of {}.'.format(refs))
        else:
            args.ref_name = list(refs)[0]
    elif args.ref_name not in refs:
        raise KeyError('Ref {} not in bam refs {}'.format(args.ref_name, refs))

    bam_files = { bam: pysam.AlignmentFile(bam) for bam in args.bams}

    if len(args.bams) > 1:  # expect just 1 read per bam
        reads = {bn: list(bam.fetch(args.ref_name)) for bn,bam in bam_files.items()}
        if not all([ len(l) == 1 for l in reads.values()]):
            raise RuntimeError('Expected just one read per bam, do not chunk and filter to primary alignments')
        reads = {bn: l[0] for bn,l in reads.items()}
    else: # expect more than 1 read
        reads = {r.query_name: r for r in bam_files[args.bams[0]].fetch()}
        if not len(reads.values()) > 1:
            raise RuntimeError('The bam {} contained just 1 read.'.format(args.bam[0]))

    print('Initial alignments:')
    for bn, read in reads.items():
        print('{} {} {} {}'.format(bn, read.query_name, read.reference_start, read.reference_end))
    start = max([r.reference_start for r in reads.values()])
    end = min([r.reference_end for r in reads.values()])
    print('Initial ref start {} ref end {}'.format(start, end))

    # trim back to point where each read is mapped to a common ref
    ref_pos = { k: read.get_reference_positions(full_length=True) for k, read in reads.items()}
    while True:
        start_in_ref_pos = [start in p for p in ref_pos.values()]
        if all(start_in_ref_pos):
            break
        else:
            start += 1
            print('Shifing forward start')

    while True:
        end_in_ref_pos = [end - 1 in p for p in ref_pos.values()]
        if all(end_in_ref_pos):
            break
        else:
            end -= 1
            print('Shifing back start')

    print('Final trimmed region: {}:{}-{}'.format(args.ref_name, start, end))

    seq_objs = []
    for bn, read in reads.items():
        ref_pos = read.get_reference_positions(full_length=True)
        q_start = ref_pos.index(start)
        q_end = ref_pos.index(end - 1) + 1
        seq = read.query_sequence[q_start:q_end]
        prefix = os.path.splitext(os.path.basename(bn))[0]
        read_id='{}_{}_{}_{}'.format(prefix, read.query_name, q_start, q_end)
        seq_ob = SeqIO.SeqRecord(Seq(seq), id=read_id)
        seq_objs.append(seq_ob)

    output = '{}_queries.fasta'.format(args.output_prefix)
    with open(output, 'w') as fh:
        SeqIO.write(seq_objs, fh, 'fasta')

    for b in bam_files.values():
        b.close()

    if args.reference_fasta is not None:  # output trimmed reference
        output_ref = '{}_reference.fasta'.format(args.output_prefix)
        ndx = SeqIO.index(args.reference_fasta, 'fasta')
        if not args.ref_name in ndx.keys():
            raise KeyError('Reference {} not in {}'.format(args.ref_name, args.fasta))

        with open(output_ref, 'w') as fh:
            ref_id='{}_{}_{}'.format(args.ref_name, start, end)
            trimmed_ref = SeqIO.SeqRecord(ndx[args.ref_name].seq[start: end], ref_id)
            SeqIO.write([trimmed_ref], fh, 'fasta')
        ndx.close()


if __name__ == '__main__':
    main()



import argparse
import logging

import pysam

from pomoxis.util import filter_args, filter_read, parse_regions, Region


def main():
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    parser = argparse.ArgumentParser(
        prog='filter_bam',
        description='Filter a bam',
        parents=[filter_args()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bam',
        help='input bam file.')
    parser.add_argument('-o', '--output_bam', default='filtered.bam',
        help='Output bam file.')
    parser.add_argument('-r', '--region',
        help='Only process given region.')
    parser.add_argument('-t', '--threads', default=1, type=int,
        help='Number of parallel threads for io processing.')

    args = parser.parse_args()
    logger = logging.getLogger('filter')

    logger.info("Starting filtering")
    in_io_threads=max(1, args.threads // 2)
    out_io_threads=max(1, args.threads - in_io_threads)
    logger.info(f"Using {in_io_threads} input and {out_io_threads} output threads for BAM compression")

    with pysam.AlignmentFile(args.bam, check_sq=False, threads=in_io_threads) as bam:
        with pysam.AlignmentFile(args.output_bam, "wb", header=bam.header, threads=out_io_threads) as out_bam:
            if args.region is not None:
                ref_lengths = dict(zip(bam.references, bam.lengths))
                region = parse_regions([args.region], ref_lengths=ref_lengths)[0]
                reads = bam.fetch(region.ref_name, region.start, region.end)
            else:
                reads = bam
            for read in reads:
                if filter_read(read, args, logger):
                    continue
                out_bam.write(read)
    pysam.index(args.output_bam, "-@", str(args.threads))
    logger.info("Finished filtering")
